# -*- coding: utf-8 -*-

"""
fhi_aims_step
A SEAMM plug-in for FHI-aims
"""

# Bring up the classes so that they appear to be directly in
# the fhi_aims_step package.

from .fhi_aims import FHIaims  # noqa: F401, E501
from .fhi_aims_parameters import FHIaimsParameters  # noqa: F401
from .fhi_aims_step import FHIaimsStep  # noqa: F401, E501
from .tk_fhi_aims import TkFHIaims  # noqa: F401, E501

from .metadata import metadata  # noqa: F401

from .energy_step import EnergyStep  # noqa: F401
from .energy import Energy  # noqa: F401
from .energy_parameters import EnergyParameters  # noqa: F401
from .tk_energy import TkEnergy  # noqa: F401

from .optimization_step import OptimizationStep  # noqa: F401
from .optimization import Optimization  # noqa: F401
from .optimization_parameters import OptimizationParameters  # noqa: F401
from .tk_optimization import TkOptimization  # noqa: F401

# Handle versioneer
from ._version import get_versions

__author__ = "Paul Saxe"
__email__ = "psaxe@molssi.org"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



# -*- coding: utf-8 -*-
"""
Control parameters for the FHIaims step in a SEAMM flowchart
"""

import logging
import seamm
import pprint  # noqa: F401

logger = logging.getLogger(__name__)


class FHIaimsParameters(seamm.Parameters):  # noqa: E999
    """
    The control parameters for FHIaims.

    You need to replace the "time" entry in dictionary below these comments with the
    definitions of parameters to control this step. The keys are parameters for the
    current plugin,the values are dictionaries as outlined below.

    Examples
    --------
    ::

        parameters = {
            "time": {
                "default": 100.0,
                "kind": "float",
                "default_units": "ps",
                "enumeration": tuple(),
                "format_string": ".1f",
                "description": "Simulation time:",
                "help_text": ("The time to simulate in the dynamics run.")
            },
        }

    parameters : {str: {str: str}}
        A dictionary containing the parameters for the current step.
        Each key of the dictionary is a dictionary that contains the
        the following keys:

    parameters["default"] :
        The default value of the parameter, used to reset it.

    parameters["kind"] : enum()
        Specifies the kind of a variable. One of  "integer", "float", "string",
        "boolean", or "enum"

        While the "kind" of a variable might be a numeric value, it may still have
        enumerated custom values meaningful to the user. For instance, if the parameter
        is a convergence criterion for an optimizer, custom values like "normal",
        "precise", etc, might be adequate. In addition, any parameter can be set to a
        variable of expression, indicated by having "$" as the first character in the
        field. For example, $OPTIMIZER_CONV.

    parameters["default_units"] : str
        The default units, used for resetting the value.

    parameters["enumeration"]: tuple
        A tuple of enumerated values.

    parameters["format_string"]: str
        A format string for "pretty" output.

    parameters["description"]: str
        A short string used as a prompt in the GUI.

    parameters["help_text"]: str
        A longer string to display as help for the user.

    See Also
    --------
    FHIaims, TkFHIaims, FHIaims FHIaimsParameters, FHIaimsStep
    """

    parameters = {
        # "time": {
        #     "default": 100.0,
        #     "kind": "float",
        #     "default_units": "ps",
        #     "enumeration": tuple(),
        #     "format_string": ".1f",
        #     "description": "Simulation time:",
        #     "help_text": ("The time to simulate in the dynamics run."),
        # },
        # # Results handling ... uncomment if needed
        # "results": {
        #     "default": {},
        #     "kind": "dictionary",
        #     "default_units": "",
        #     "enumeration": tuple(),
        #     "format_string": "",
        #     "description": "results",
        #     "help_text": "The results to save to variables or in tables.",
        # },
    }

    def __init__(self, defaults={}, data=None):
        """
        Initialize the parameters, by default with the parameters defined above

        Parameters
        ----------
        defaults: dict
            A dictionary of parameters to initialize. The parameters
            above are used first and any given will override/add to them.
        data: dict
            A dictionary of keys and a subdictionary with value and units
            for updating the current, default values.

        Returns
        -------
        None
        """

        logger.debug("FHIaimsParameters.__init__")

        super().__init__(
            defaults={**FHIaimsParameters.parameters, **defaults}, data=data
        )



# -*- coding: utf-8 -*-
"""
Control parameters for the Energy step in a SEAMM flowchart
"""

import logging
import seamm
import pprint  # noqa: F401

logger = logging.getLogger(__name__)


class EnergyParameters(seamm.Parameters):  # noqa: E999
    """
    The control parameters for Energy.

    You need to replace the "time" entry in dictionary below these comments with the
    definitions of parameters to control this step. The keys are parameters for the
    current plugin,the values are dictionaries as outlined below.

    Examples
    --------
    ::

        parameters = {
            "time": {
                "default": 100.0,
                "kind": "float",
                "default_units": "ps",
                "enumeration": tuple(),
                "format_string": ".1f",
                "description": "Simulation time:",
                "help_text": ("The time to simulate in the dynamics run.")
            },
        }

    parameters : {str: {str: str}}
        A dictionary containing the parameters for the current step.
        Each key of the dictionary is a dictionary that contains the
        the following keys:

    parameters["default"] :
        The default value of the parameter, used to reset it.

    parameters["kind"] : enum()
        Specifies the kind of a variable. One of  "integer", "float", "string",
        "boolean", or "enum"

        While the "kind" of a variable might be a numeric value, it may still have
        enumerated custom values meaningful to the user. For instance, if the parameter
        is a convergence criterion for an optimizer, custom values like "normal",
        "precise", etc, might be adequate. In addition, any parameter can be set to a
        variable of expression, indicated by having "$" as the first character in the
        field. For example, $OPTIMIZER_CONV.

    parameters["default_units"] : str
        The default units, used for resetting the value.

    parameters["enumeration"]: tuple
        A tuple of enumerated values.

    parameters["format_string"]: str
        A format string for "pretty" output.

    parameters["description"]: str
        A short string used as a prompt in the GUI.

    parameters["help_text"]: str
        A longer string to display as help for the user.

    See Also
    --------
    Energy, TkEnergy, Energy EnergyParameters, EnergyStep
    """

    parameters = {
        "input only": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": (
                "yes",
                "no",
            ),
            "format_string": "s",
            "description": "Write the input files and stop:",
            "help_text": "Don't run MOPAC. Just write the input files.",
        },
        "gui": {
            "default": (
                "with recommended features for energy or structure of molecules "
                "or nonmetallic solids"
            ),
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                (
                    "with recommended features for energy or structure involving "
                    "metallic solids"
                ),
                (
                    "with recommended features for energy or structure of molecules "
                    "or nonmetallic solids"
                ),
                "with recommended features for band structures of metallic solids",
                (
                    "with recommended features for orbital energies of molecules or "
                    "band structures for nonmetallic solids"
                ),
                "with all standard features",
                "including experimental features",
            ),
            "format_string": "",
            "description": "Configure GUI",
            "help_text": "How to configure the GUI",
        },
        "calculate_gradients": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Calculate gradients:",
            "help_text": (
                "Whether to calculate the gradients in a single-point calculation."
            ),
        },
        "results": {
            "default": {},
            "kind": "dictionary",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "results",
            "help_text": "The results to save to variables or in tables.",
        },
    }

    energy_parameters = {
        "basis_version": {
            "default": "defaults_2020",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "defaults_2010",
                "defaults_2020",
                "NAO-VCC-nZ",
                "NAO-J-n",
                "non-standard/gaussian_tight_770",
                "non-standard/Tier2_aug2",
                "custom",
            ),
            "format_string": "",
            "description": "Basis set version:",
            "help_text": "The version of basis sets used.",
        },
        "defaults_2010_level": {
            "default": "light",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "light",
                "intermediate",
                "tight",
                "really_tight",
            ),
            "format_string": "",
            "description": "Basis set level:",
            "help_text": "The level of 2010 basis sets used.",
        },
        "defaults_2020_level": {
            "default": "light",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "light_spd/light",
                "light",
                "intermediate",
                "tight",
                "really_tight",
            ),
            "format_string": "",
            "description": "Basis set level:",
            "help_text": "The level of 2010 basis sets used.",
        },
        "NAO-VCC-nZ_level": {
            "default": "NAO-VCC-2Z",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "NAO-VCC-2Z",
                "NAO-VCC-3Z",
                "NAO-VCC-4Z",
                "NAO-VCC-5Z",
            ),
            "format_string": "",
            "description": "Basis set level:",
            "help_text": "The level of NAO-VCC (valence correlation) basis sets used.",
        },
        "NAO-J-n_level": {
            "default": "NAO-J-2",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "NAO-J-2",
                "NAO-J-3",
                "NAO-J-4",
                "NAO-J-5",
            ),
            "format_string": "",
            "description": "Basis set level:",
            "help_text": "The level of NAO-j (spin-coupling) basis sets used.",
        },
        "model": {
            "default": "Generalized-gradient approximations (GGA)",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "Local-density approximation (LDA)",
                "Generalized-gradient approximations (GGA)",
                "Meta-generalized gradient approximations (meta-GGA)",
                "Hartree-Fock and hybrid functionals",
                "Hybrid meta generalized gradient approximation (Hybrid meta-GGA)",
            ),
            "format_string": "",
            "description": "Computational model:",
            "help_text": "The high-level computational model employed.",
        },
        "submodel": {
            "default": "pbe : Perdew, Burke and Ernzerhof",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "pbe : Perdew, Burke and Ernzerhof",
                "am05 : Armiento and Mattsson",
                "blyp : Becke and Lee-Yang-Parr",
                "pbeint",
                "pbesol : modified PBE GGA",
                "rpbe : the RPBE modified PBE functional",
                "revpbe : the revPBE modified PBE GGA",
                "r48pbe : 0.52*pbe and 0.48*rpbe",
                "pw91_gga : Perdew-Wang 1991 GGA",
                "b86bbpe : Becke’s B86b + PBE",
            ),
            "format_string": "",
            "description": "Model version:",
            "help_text": "The version of the model employed.",
        },
        "spin_polarization": {
            "default": "default",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ("default", "none"),
            "format_string": "",
            "description": "Spin polarized?:",
            "help_text": (
                "Whether to do a spin polarized calculation. The default is to do so "
                "if there are spins on any atoms. You can also turn this off, but "
                "cannot force a spin-polarized calculation without setting spins on "
                "one or more atoms in the structure."
            ),
        },
        "fixed_spin_moment": {
            "default": "no",
            "kind": "integer",
            "default_units": "",
            "enumeration": ("no", "yes"),
            "format_string": "",
            "description": "Fix the spin moment:",
            "help_text": (
                "Whether to fix the spin moment, by default to the configuration's "
                "multiplicity. You may also specifiy and integer multiplicity here."
            ),
        },
        "relativity": {
            "default": "atomic ZORA approximation",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ("none", "atomic ZORA approximation"),
            "format_string": "",
            "description": "Relativity:",
            "help_text": "Whether and how to calculate relatavistic effects.",
        },
        "dispersion": {
            "default": "nonlocal many-body dispersion (MBD-NL)",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "none",
                "nonlocal many-body dispersion (MBD-NL)",
                "range-separated self-consistently screened (MBD@rsSCS)",
            ),
            "format_string": "",
            "description": "Long-range vdW term:",
            "help_text": (
                "Whether and how to calculate the many-body dispersions effects."
            ),
        },
        "primitive cell": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "If possible, use primitive cell:",
            "help_text": "Whether to use the primitive cell for the calculation.",
        },
    }

    kspace_parameters = {
        "k-grid method": {
            "default": "grid spacing",
            "kind": "string",
            "default_units": "",
            "enumeration": ("grid spacing", "explicit grid dimensions"),
            "format_string": "",
            "description": "Specify k-space grid using:",
            "help_text": "How to specify the k-space integration grid.",
        },
        "k-spacing": {
            "default": 0.2,
            "kind": "float",
            "default_units": "1/Å",
            "enumeration": None,
            "format_string": "",
            "description": "K-spacing:",
            "help_text": "The spacing of the grid in reciprocal space.",
        },
        "odd grid": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Force to odd numbers:",
            "help_text": "Whether to force the grid sizes to odd numbers.",
        },
        "centering": {
            "default": "𝚪-centered",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "𝚪-centered",
                "off-center",
                "Monkhorst-Pack",
            ),
            "format_string": "",
            "description": "Centering of grid:",
            "help_text": "How to center the grid in reciprocal space.",
        },
        "occupation type": {
            "default": "Gaussian",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "Gaussian",
                "Methfessel-Paxton",
                "Fermi",
                "integer",
                "cubic",
                "cold",
            ),
            "format_string": "",
            "description": "Smearing:",
            "help_text": (
                "How occupy the orbitals, typically smearing the electrons as they "
                "would be at finite temperature."
            ),
        },
        "smearing width": {
            "default": 0.01,
            "kind": "float",
            "default_units": "eV",
            "enumeration": None,
            "format_string": ".3f",
            "description": "Width:",
            "help_text": "The smearing or broadening width.",
        },
        "na": {
            "default": 4,
            "kind": "integer",
            "default_units": "",
            "enumeration": None,
            "format_string": "",
            "description": "NPoints in a:",
            "help_text": (
                "Number of points in the first direction of the Brillouin zone."
            ),
        },
        "nb": {
            "default": 4,
            "kind": "integer",
            "default_units": "",
            "enumeration": None,
            "format_string": "",
            "description": "b:",
            "help_text": (
                "Number of points in the second direction of the Brillouin zone."
            ),
        },
        "nc": {
            "default": 4,
            "kind": "integer",
            "default_units": "",
            "enumeration": None,
            "format_string": "",
            "description": "c:",
            "help_text": (
                "Number of points in the third direction of the Brillouin zone."
            ),
        },
    }

    output_parameters = {
        "total density": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot total density:",
            "help_text": "Whether to plot the total charge density.",
        },
        "total spin density": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot total spin density:",
            "help_text": "Whether to plot the total spin density.",
        },
        "difference density": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot difference density:",
            "help_text": "Whether to plot the difference density.",
        },
        "orbitals": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Plot orbitals:",
            "help_text": "Whether to plot orbitals.",
        },
        "selected orbitals": {
            "default": "-1, HOMO, LUMO, +1",
            "kind": "string",
            "default_units": "",
            "enumeration": ("all", "-1, HOMO, LUMO, +1"),
            "format_string": "",
            "description": "Selected orbitals:",
            "help_text": "Which orbitals to plot.",
        },
        "selected k-points": {
            "default": "none",
            "kind": "string",
            "default_units": "",
            "enumeration": ("none", "all"),
            "format_string": "",
            "description": "For crystals, k-points:",
            "help_text": "Plots the orbitals at these k-points.",
        },
        "region": {
            "default": "default",
            "kind": "string",
            "default_units": "",
            "enumeration": ("default", "explicit"),
            "format_string": "",
            "description": "Region:",
            "help_text": "The region for the plots",
        },
        "nx": {
            "default": 50,
            "kind": "integer",
            "default_units": "",
            "enumeration": None,
            "format_string": "",
            "description": "Grid:",
            "help_text": "Number of grid points in first direction",
        },
        "ny": {
            "default": 50,
            "kind": "integer",
            "default_units": "",
            "enumeration": None,
            "format_string": "",
            "description": "x",
            "help_text": "Number of grid points in second direction",
        },
        "nz": {
            "default": 50,
            "kind": "integer",
            "default_units": "",
            "enumeration": None,
            "format_string": "",
            "description": "x",
            "help_text": "Number of grid points in first direction",
        },
    }

    def __init__(self, defaults={}, data=None):
        """
        Initialize the parameters, by default with the parameters defined above

        Parameters
        ----------
        defaults: dict
            A dictionary of parameters to initialize. The parameters
            above are used first and any given will override/add to them.
        data: dict
            A dictionary of keys and a subdictionary with value and units
            for updating the current, default values.

        Returns
        -------
        None
        """

        logger.debug("EnergyParameters.__init__")

        super().__init__(
            defaults={
                **EnergyParameters.parameters,
                **EnergyParameters.energy_parameters,
                **EnergyParameters.kspace_parameters,
                **EnergyParameters.output_parameters,
                **defaults,
            },
            data=data,
        )



# -*- coding: utf-8 -*-

"""The graphical part of a FHI-aims step"""

import pprint  # noqa: F401

import fhi_aims_step  # noqa: F401
import seamm
from seamm_util import ureg, Q_, units_class  # noqa: F401


class TkFHIaims(seamm.TkNode):
    """
    The graphical part of a FHI-aims step in a flowchart.

    Attributes
    ----------
    tk_flowchart : TkFlowchart = None
        The flowchart that we belong to.
    node : Node = None
        The corresponding node of the non-graphical flowchart
    canvas: tkCanvas = None
        The Tk Canvas to draw on
    dialog : Dialog
        The Pmw dialog object
    x : int = None
        The x-coordinate of the center of the picture of the node
    y : int = None
        The y-coordinate of the center of the picture of the node
    w : int = 200
        The width in pixels of the picture of the node
    h : int = 50
        The height in pixels of the picture of the node
    self[widget] : dict
        A dictionary of tk widgets built using the information
        contained in FHI-aims_parameters.py

    See Also
    --------
    FHIaims, TkFHIaims,
    FHIaimsParameters,
    """

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        namespace="org.molssi.seamm.fhi_aims.tk",
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
    ):
        """
        Initialize a graphical node.

        Parameters
        ----------
        tk_flowchart: Tk_Flowchart
            The graphical flowchart that we are in.
        node: Node
            The non-graphical node for this step.
        namespace: str
            The stevedore namespace for finding sub-nodes.
        canvas: Canvas
           The Tk canvas to draw on.
        x: float
            The x position of the nodes center on the canvas.
        y: float
            The y position of the nodes cetner on the canvas.
        w: float
            The nodes graphical width, in pixels.
        h: float
            The nodes graphical height, in pixels.

        Returns
        -------
        None
        """
        self.namespace = namespace
        self.dialog = None

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
        )
        self.create_dialog()

    def create_dialog(self):
        """
        Create the dialog. A set of widgets will be chosen by default
        based on what is specified in the FHI-aims_parameters
        module.

        Parameters
        ----------
        None

        Returns
        -------
        None

        See Also
        --------
        TkFHIaims.reset_dialog
        """

        frame = super().create_dialog(title="FHI-aims")
        # make it large!
        screen_w = self.dialog.winfo_screenwidth()
        screen_h = self.dialog.winfo_screenheight()
        w = int(0.9 * screen_w)
        h = int(0.8 * screen_h)
        x = int(0.05 * screen_w / 2)
        y = int(0.1 * screen_h / 2)

        self.dialog.geometry(f"{w}x{h}+{x}+{y}")

        self.tk_subflowchart = seamm.TkFlowchart(
            master=frame, flowchart=self.node.subflowchart, namespace=self.namespace
        )
        self.tk_subflowchart.draw()

    def right_click(self, event):
        """
        Handles the right click event on the node.

        Parameters
        ----------
        event : Tk Event

        Returns
        -------
        None

        See Also
        --------
        TkFHIaims.edit
        """

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)



# -*- coding: utf-8 -*-

"""This file contains metadata describing the results from FHIaims
"""

metadata = {}

"""Description of the computational models for FHIaims.

Hamiltonians, approximations, and basis set or parameterizations,
only if appropriate for this code. For example::

    metadata["computational models"] = {
        "Hartree-Fock": {
            "models": {
                "PM7": {
                    "parameterizations": {
                        "PM7": {
                            "elements": "1-60,62-83",
                            "periodic": True,
                            "reactions": True,
                            "optimization": True,
                            "code": "mopac",
                        },
                        "PM7-TS": {
                            "elements": "1-60,62-83",
                            "periodic": True,
                            "reactions": True,
                            "optimization": False,
                            "code": "mopac",
                        },
                    },
                },
            },
        },
    }
"""

metadata["computational models"] = {
    "Density Functional Theory (DFT)": {
        "models": {
            "Local-density approximation (LDA)": {
                "parameterizations": {
                    "vwn : Vosko-Wilk-Nusair": {
                        "gui": "metal/energy",
                        "description": "LDA of Vosko, Wilk, and Nusair 1980 [233].",
                    },
                    "pw-lda : Perdew-Wang": {
                        "gui": "metal/energy",
                        "description": (
                            "Homogeneous electron gas based on Ceperley and Alder [42] "
                            "as parameterized by Perdew and Wang 1992 [183]. "
                            "Recommended LDA parameterization."
                        ),
                    },
                    "pz-lda : Perdew-Zunger": {
                        "gui": "metal/energy",
                        "description": (
                            "Homogeneous electron gas based on Ceperley and Alder [42],"
                            " as parameterized by Perdew and Zunger 1981 [184]"
                        ),
                    },
                    "vwn-gauss : VWN based on random phase approx.": {
                        "gui": "standard",
                        "description": (
                            "LDA of Vosko, Wilk, and Nusair 1980, but based on the "
                            "random phase approximation [210]. Do not use this LDA "
                            "unless for one specific reason: In the B3LYP "
                            "implementation of the Gaussian code, this functional "
                            "is allegedly used instead of the correct VWN functional. "
                            "It is therefore now present in many reference results in "
                            "the literature, and also available here for comparison."
                        ),
                    },
                }
            },
            "Generalized-gradient approximations (GGA)": {
                "parameterizations": {
                    "pbe : Perdew, Burke and Ernzerhof": {
                        "gui": "metal/energy molecule/energy",
                        "description": "GGA of Perdew, Burke and Ernzerhof 1997 [181].",
                    },
                    "am05 : Armiento and Mattsson": {
                        "gui": "standard",
                        "description": (
                            "GGA functional designed to include surface effects in "
                            "self-consistent density functional theory, according to "
                            "Armiento and Mattsson [9]"
                        ),
                    },
                    "blyp : Becke and Lee-Yang-Parr": {
                        "gui": "standard",
                        "description": (
                            "The BLYP functional: Becke (1988) exchange [20] and "
                            "Lee-Yang-Parr correlation [150]."
                        ),
                    },
                    "pbeint": {
                        "gui": "standard",
                        "description": "PBEint functional of Ref. [67]",
                    },
                    "pbesol : modified PBE GGA": {
                        "gui": "metal/energy",
                        "description": "Modified PBE GGA according to Ref. [187].",
                    },
                    "rpbe : the RPBE modified PBE functional": {
                        "gui": "standard",
                        "description": (
                            "The RPBE modified PBE functional according to Ref. [93]."
                        ),
                    },
                    "revpbe : the revPBE modified PBE GGA": {
                        "gui": "standard",
                        "description": (
                            "The revPBE modified PBE GGA suggested in Ref. [247]."
                        ),
                    },
                    "r48pbe : 0.52*pbe and 0.48*rpbe": {
                        "gui": "standard",
                        "description": (
                            "The mixed functional containing 0.52*pbe and 0.48*rpbe "
                            "according to Ref. [173]"
                        ),
                    },
                    "pw91_gga : Perdew-Wang 1991 GGA": {
                        "gui": "metal/energy",
                        "description": (
                            "GGA according to Perdew and Wang, usually referred to as "
                            "'Perdew-Wang 1991 GGA'. This GGA is most accessibly "
                            "described in Reference 26 and 27 of Ref. [182]. Note "
                            "that the often mis-quoted reference [183] does not(!) "
                            "describe the Perdew-Wang GGA but instead only the "
                            "correlation part of the local-density approximation "
                            "described above."
                        ),
                    },
                    "b86bbpe : Becke’s B86b + PBE": {
                        "gui": "molecule/energy",
                        "description": (
                            "Becke’s B86b exchange[19] plus PBE correlation[181]."
                        ),
                    },
                },
            },
            "Meta-generalized gradient approximations (meta-GGA)": {
                "parameterizations": {
                    "m06-l : Truhlar’s optimized meta-GGA": {
                        "gui": "recommended",
                        "description": (
                            "Truhlar’s optimized meta-GGA of the 'M06' suite of "
                            "functionals. [250]"
                        ),
                    },
                    "m11-l : Truhlar’s optimized range-separated local meta-GGA": {
                        "gui": "recommended",
                        "description": (
                            "Truhlar’s optimized range-separated local meta-GGA of the "
                            "'M11' suite of functionals. [191]"
                        ),
                    },
                    "revtpss": {
                        "gui": "standard",
                        "description": (
                            "Meta-GGA revTPSS functional of Ref. [185, 186]",
                        ),
                    },
                    "tpss": {
                        "gui": "recommended",
                        "description": "Meta-GGA TPSS functional of Ref. [223]",
                    },
                    "tpssloc": {
                        "gui": "standard",
                        "description": (
                            "Meta-GGA TPSSloc functional, thanks to E. Fabiano and "
                            "F. Della Sala. L.A. Constantin, E. Fabiano, F.Della Sala, "
                            "Ref. [51]."
                        ),
                    },
                    "scan or SCAN:": {
                        "gui": "recommended",
                        "description": (
                            "'Strongly Constrained and Appropriately Normed Semilocal "
                            "Density Functional,' i.e., the SCAN meta-GGA functional "
                            "by Sun, Ruzsinszky, and Perdew.[219]"
                        ),
                    },
                },
            },
            "Hybrid functionals and Hartree-Fock": {
                "parameterizations": {
                    "b3lyp": {
                        "gui": "molecule/energy",
                        "description": (
                            "'B3LYP' hybrid functional as allegedly implemented in the "
                            "Gaussian code (i.e., using the RPA version of the "
                            "Vosko-Wilk-Nusair local-density approximation, see Refs. "
                            "[233, 210] for details). Note that this is therefore not "
                            "exactly the same B3LYP as originally described by Becke "
                            "in 1993."
                        ),
                    },
                    "hf : Hartree-Fock": {
                        "gui": "standard",
                        "description": "Hartree-Fock exchange, only",
                    },
                    "hse03 : Heyd, Scuseria and Ernzerhof": {
                        "gui": "standard",
                        "description": (
                            "Hybrid functional as used in Heyd, Scuseria and Ernzerhof "
                            "[105, 106]. In this functional, 25 % of the exchange "
                            "energy is split into a short-ranged, screened Hartree-Fock"
                            " part, and a PBE GGA-like functional for the longrange "
                            "part of exchange. The remaining 75 % exchange and full "
                            "correlation energy are treated as in PBE. As clarified in "
                            "Refs. [141, 106], two different screening parameters were "
                            "used in the short-range exchange part and longrange "
                            "exchange part of the original HSE functional, "
                            "respectively. The ’hse03’ functional in FHI-aims "
                            "reproduces these original values exactly."
                        ),
                    },
                    "hse06 : Heyd, Scuseria and Ernzerhof": {
                        "gui": "molecule/energy",
                        "description": (
                            "Hybrid functional according to Heyd, Scuseria and "
                            "Ernzerhof [105], following the naming convention "
                            "suggested in Ref. [141]. In this case, the additional "
                            "option value is needed, representing the single real, "
                            "positive screening parameter omega as clarified in Ref. "
                            "[141]. In this functional, 25 % of the exchange energy is "
                            "split into a short-ranged, screened Hartree-Fock part, "
                            "and a PBE GGA-like functional for the long-range part of "
                            "exchange. The remaining 75 % exchange and full "
                            "correlation energy are treated as in PBE."
                        ),
                    },
                    "pbe0": {
                        "gui": "molecule/energy",
                        "description": (
                            "PBE0 hybrid functional [1], mixing 75 % GGA exchange with "
                            "25 % Hartree-Fock exchange"
                        ),
                    },
                    "pbesol0": {
                        "gui": "recommended",
                        "description": (
                            "Hybrid functional in analogy to PBE0 [1], except that the "
                            "PBEsol [187] GGA functionals are used, mixing 75 % GGA "
                            "exchange with 25 % Hartree-Fock exchange."
                        ),
                    },
                    "b86bpbe-25": {
                        "gui": "molecule/energy",
                        "description": (
                            "B86bPBE hybrid functional [19, 181], mixing 75 % GGA "
                            "exchange with 25 % Hartree-Fock exchange."
                        ),
                    },
                    "b86bpbe-50": {
                        "gui": "standard",
                        "description": (
                            "B86bPBE hybrid functional [19, 181], mixing 75 % GGA "
                            "exchange with 50 % Hartree-Fock exchange."
                        ),
                    },
                    "lc_wpbeh": {
                        "gui": "standard",
                        "description": (
                            "Range separated hybrid functional LC-!PBEh using 100 % "
                            "Hartree-Fock exchange in the long-range part and omegaPBE "
                            "[179] in the shortrange part. The full correlation energy "
                            "is treated as in PBE."
                        ),
                    },
                },
            },
            "Hybrid meta generalized gradient approximation (Hybrid meta-GGA)": {
                "parameterizations": {
                    "m06": {
                        "gui": "recommended",
                        "description": (
                            "Truhlar’s optimized hybrid meta-GGA of the 'M06' suite of "
                            "functionals; with 27% exact exchange. [249]"
                        ),
                    },
                    "m06-2x": {
                        "gui": "standard",
                        "description": (
                            "Truhlar’s optimized hybrid meta-GGA of the 'M06' suite of "
                            "functionals, with double contribution (54%) from the "
                            "hartree-fock exact exchange. [249]"
                        ),
                    },
                    "m06-hf": {
                        "gui": "standard",
                        "description": (
                            "Truhlar’s optimized hybrid meta-GGA of the 'M06' suite of "
                            "functionals, with 100% exact exchage contribution. [248]"
                        ),
                    },
                    "m08-hx": {
                        "gui": "recommended",
                        "description": (
                            "Truhlar’s optimized hybrid meta-GGA of the 'M08' suite of "
                            "functionals, with 52.23% contribution from the "
                            "hartree-fock exact exchange. [251]"
                        ),
                    },
                    "m08-so": {
                        "gui": "standard",
                        "description": (
                            "Truhlar’s optimized hybrid meta-GGA of the 'M08' suite of "
                            "functionals, with 56.79% contribution from the "
                            "hartree-fock exact exchange. [251]"
                        ),
                    },
                    "m11": {
                        "gui": "recommended",
                        "description": (
                            "Truhlar’s optimized range-separated local meta-GGA of the "
                            "'M11' suite of functionals [190]. The range-separation "
                            "variable is also hardcoded in this implementation with "
                            "omega = 0.25 bohr−1"
                        ),
                    },
                },
            },
            # "": {
            #     "parameterizations": {
            #         "": {
            #             "description": (
            #             ),
            #         },
            #         "": {
            #             "description": (
            #             ),
            #         },
            #         "": {
            #             "description": (
            #             ),
            #         },
            #         "": {
            #             "description": (
            #             ),
            #         },
            #         "": {
            #             "description": (
            #             ),
            #         },
            #     },
            # },
        }
    }
}

"""Description of the FHIaims keywords.

(Only needed if this code uses keywords)

Fields
------
description : str
    A human readable description of the keyword.
takes values : int (optional)
    Number of values the keyword takes. If missing the keyword takes no values.
default : str (optional)
    The default value(s) if the keyword takes values.
format : str (optional)
    How the keyword is formatted in the MOPAC input.

For example::
    metadata["keywords"] = {
        "0SCF": {
            "description": "Read in data, then stop",
        },
        "ALT_A": {
            "description": "In PDB files with alternative atoms, select atoms A",
            "takes values": 1,
            "default": "A",
            "format": "{}={}",
        },
    }
"""
# metadata["keywords"] = {
# }

"""Properties that FHIaims produces.
`metadata["results"]` describes the results that this step can produce. It is a
dictionary where the keys are the internal names of the results within this step, and
the values are a dictionary describing the result. For example::

    metadata["results"] = {
        "total_energy": {
            "calculation": [
                "energy",
                "optimization",
            ],
            "description": "The total energy",
            "dimensionality": "scalar",
            "methods": [
                "ccsd",
                "ccsd(t)",
                "dft",
                "hf",
            ],
            "property": "total energy#Psi4#{model}",
            "type": "float",
            "units": "E_h",
        },
    }

Fields
______

calculation : [str]
    Optional metadata describing what subtype of the step produces this result.
    The subtypes are completely arbitrary, but often they are types of calculations
    which is why this is name `calculation`. To use this, the step or a substep
    define `self._calculation` as a value. That value is used to select only the
    results with that value in this field.

description : str
    A human-readable description of the result.

dimensionality : str
    The dimensions of the data. The value can be "scalar" or an array definition
    of the form "[dim1, dim2,...]". Symmetric tringular matrices are denoted
    "triangular[n,n]". The dimensions can be integers, other scalar
    results, or standard parameters such as `n_atoms`. For example, '[3]',
    [3, n_atoms], or "triangular[n_aos, n_aos]".

methods : str
    Optional metadata like the `calculation` data. `methods` provides a second
    level of filtering, often used for the Hamiltionian for *ab initio* calculations
    where some properties may or may not be calculated depending on the type of
    theory.

property : str
    An optional definition of the property for storing this result. Must be one of
    the standard properties defined either in SEAMM or in this steps property
    metadata in `data/properties.csv`.

type : str
    The type of the data: string, integer, or float.

units : str
    Optional units for the result. If present, the value should be in these units.
"""
metadata["results"] = {
    "total_energy": {
        "calculation": [
            "energy",
            "optimization",
        ],
        "description": "The total energy",
        "dimensionality": "scalar",
        "property": "total energy#FHIaims#{model}",
        "type": float,
        "units": "eV",
        "json_section": "final_output",
        "re": (
            r"Total energy of the DFT / Hartree-Fock s.c.f. calculation      :"
            r" *([-+.0-9]+) eV"
        ),
    },
    "dispersion_energy": {
        "calculation": [
            "energy",
            "optimization",
        ],
        "description": "vdW energy correction",
        "dimensionality": "scalar",
        "type": float,
        "units": "eV",
        "re": r"Libmbd: Evaluated energy: *([-+E.0-9]+)",
    },
    "total_number_of_loops": {
        "calculation": [
            "energy",
            "optimization",
        ],
        "description": "Number of SCF iterations",
        "dimensionality": "scalar",
        "type": int,
        "json_section": "final_output",
    },
    "relaxation_step_number": {
        "calculation": [
            "optimization",
        ],
        "description": "Number of geometry steps",
        "dimensionality": "scalar",
        "type": int,
        "json_section": "final_output",
    },
    "maximum_force": {
        "calculation": [
            "optimization",
        ],
        "description": "Maximum force",
        "dimensionality": "scalar",
        "type": float,
        "units": "eV/Å",
        "re": r"Maximum force component is *([-+E.0-9]+) +eV/A",
    },
    "norm_force_atoms": {
        "calculation": [
            "optimization",
        ],
        "description": "Norm of force on atoms",
        "dimensionality": "scalar",
        "type": float,
        "units": "eV/Å",
        "re": r"\|\| Forces on atoms   \|\| = *([-+E.0-9]+) +eV/A",
    },
    "n_basis": {
        "description": "Number of basis fnctns",
        "dimensionality": "scalar",
        "type": int,
        "json_section": "final_output",
    },
    "n_electrons": {
        "description": "Number of electrons",
        "dimensionality": "scalar",
        "type": float,
        "json_section": "final_output",
    },
    "n_spin": {
        "description": "Number of spins",
        "dimensionality": "scalar",
        "type": int,
        "json_section": "final_output",
    },
    "time_total": {
        "description": "Total cpu time",
        "dimensionality": "scalar",
        "type": float,
        "json_section": "final_output",
    },
    "clock_time_total": {
        "description": "Total wallclock time",
        "dimensionality": "scalar",
        "type": float,
        "json_section": "final_output",
    },
    "total_energy^": {
        "calculation": [
            "energy",
            "optimization",
        ],
        "description": "SCF iteration energy",
        "dimensionality": "[n_iterations]",
        "type": float,
        "units": "eV",
        "json_section": "scf_iteration",
    },
    "change_charge_density": {
        "description": "SCF final delta charge density",
        "dimensionality": "[n_iterations]",
        "type": float,
        "json_section": "scf_iteration",
    },
    "change_spin_density": {
        "description": "SCF final delta spin density",
        "dimensionality": "[n_iterations]",
        "type": float,
        "json_section": "scf_iteration",
    },
    "change_sum_eigenvalues": {
        "description": "SCF final delta eigenvalue sum",
        "dimensionality": "[n_iterations]",
        "type": float,
        "json_section": "scf_iteration",
    },
    "change_forces": {
        "description": "SCF final delta forces",
        "dimensionality": "[n_iterations]",
        "type": float,
        "json_section": "scf_iteration",
    },
    "atoms_proj_charge": {
        "description": "Mulliken charges on atoms",
        "dimensionality": "[n_atoms]",
        "type": float,
        "json_section": "mulliken",
    },
    "atoms_proj_spin": {
        "description": "Mulliken spins on atoms",
        "dimensionality": "[n_atoms]",
        "type": float,
        "json_section": "mulliken",
    },
    "total_spin": {
        "description": "Total spin",
        "dimensionality": "scalar",
        "type": float,
        "json_section": "mulliken",
    },
}



# -*- coding: utf-8 -*-

"""The graphical part of a Optimization step"""

import logging
import pprint  # noqa: F401
import tkinter as tk
import tkinter.ttk as ttk

import fhi_aims_step
from seamm_util import ureg, Q_, units_class  # noqa: F401, E999
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkOptimization(fhi_aims_step.TkEnergy):
    """
    The graphical part of a Optimization step in a flowchart.

    Attributes
    ----------
    tk_flowchart : TkFlowchart = None
        The flowchart that we belong to.
    node : Node = None
        The corresponding node of the non-graphical flowchart
    canvas: tkCanvas = None
        The Tk Canvas to draw on
    dialog : Dialog
        The Pmw dialog object
    x : int = None
        The x-coordinate of the center of the picture of the node
    y : int = None
        The y-coordinate of the center of the picture of the node
    w : int = 200
        The width in pixels of the picture of the node
    h : int = 50
        The height in pixels of the picture of the node
    self[widget] : dict
        A dictionary of tk widgets built using the information
        contained in Optimization_parameters.py

    See Also
    --------
    Optimization, TkOptimization,
    OptimizationParameters,
    """

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """
        Initialize a graphical node.

        Parameters
        ----------
        tk_flowchart: Tk_Flowchart
            The graphical flowchart that we are in.
        node: Node
            The non-graphical node for this step.
        namespace: str
            The stevedore namespace for finding sub-nodes.
        canvas: Canvas
           The Tk canvas to draw on.
        x: float
            The x position of the nodes center on the canvas.
        y: float
            The y position of the nodes cetner on the canvas.
        w: float
            The nodes graphical width, in pixels.
        h: float
            The nodes graphical height, in pixels.

        Returns
        -------
        None
        """
        self.dialog = None

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

    def create_dialog(self, title="Optimization"):
        """
        Create the dialog. A set of widgets will be chosen by default
        based on what is specified in the Optimization_parameters
        module.

        Parameters
        ----------
        None

        Returns
        -------
        None

        See Also
        --------
        TkOptimization.reset_dialog
        """

        super().create_dialog(title=title)

        # Shortcut for parameters
        P = self.node.parameters

        # Frame to isolate widgets
        opt_frame = self["optimization frame"] = ttk.LabelFrame(
            self["main frame"],
            borderwidth=4,
            relief="sunken",
            text="Optimization Parameters",
            labelanchor="n",
            padding=10,
        )

        for key in fhi_aims_step.OptimizationParameters.parameters:
            self[key] = P[key].widget(opt_frame)

        # Create the structure-handling widgets
        sframe = self["structure frame"] = ttk.LabelFrame(
            self["main frame"], text="Configuration Handling", labelanchor=tk.N
        )
        row = 0
        widgets = []
        for key in ("structure handling", "system name", "configuration name"):
            self[key] = P[key].widget(sframe)
            self[key].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self[key])
            row += 1
        sw.align_labels(widgets, sticky=tk.E)

        # and lay them out
        self.reset_dialog()

    def reset_dialog(self, widget=None):
        """Layout the widgets in the dialog.

        The widgets are chosen by default from the information in
        Optimization_parameter.

        This function simply lays them out row by row with
        aligned labels. You may wish a more complicated layout that
        is controlled by values of some of the control parameters.
        If so, edit or override this method

        Parameters
        ----------
        widget : Tk Widget = None

        Returns
        -------
        None

        See Also
        --------
        TkOptimization.create_dialog
        """

        row = super().reset_dialog()

        self["optimization frame"].grid(row=row, column=0, sticky=tk.N, pady=5)
        self["structure frame"].grid(
            row=row, column=1, columnspan=2, sticky=tk.N, pady=5
        )
        row += 1

        self.reset_optimization_frame()

        return row

    def reset_optimization_frame(self):
        """Layout the optimization frame according to the current values."""
        frame = self["optimization frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        widgets = []
        row = 0

        for key in ("force_convergence", "optimize_cell", "pressure"):
            w = self[key]
            w.grid(row=row, column=0, columnspan=3, sticky=tk.W)
            widgets.append(w)
            row += 1

        sw.align_labels(widgets, sticky=tk.E)

        frame.columnconfigure(0, minsize=50)

        return row



# -*- coding: utf-8 -*-

import fhi_aims_step  # noqa: E999


class EnergyStep(object):
    """Helper class needed for the stevedore integration.

    This must provide a `description()` method that returns a dict containing a
    description of this node, and `create_node()` and `create_tk_node()` methods
    for creating the graphical and non-graphical nodes.

    The dictionary for the description is the class variable just below these
    comments. The felds are as follows:

        my_description : {str, str}
            A human-readable description of this step. It can be
            several lines long, and needs to be clear to non-expert users.
            It contains the following keys: description, group, name.

        my_description["description"] : tuple
            A description of the Energy step. It must be
            clear to non-experts.

        my_description["group"] : str
            Which group in the menus to put this step. If the group does
            not exist it will be created. Common groups are "Building",
            "Control", "Custom", "Data", and "Simulations".

        my_description["name"] : str
            The name of this step, to be displayed in the menus.
    """

    my_description = {
        "description": "An interface for Energy",
        "group": "Calculations",
        "name": "Energy",
    }

    def __init__(self, flowchart=None, gui=None):
        """Initialize this helper class, which is used by
        the application via stevedore to get information about
        and create node objects for the flowchart
        """
        pass

    def create_node(self, flowchart=None, **kwargs):
        """Create and return the new node object.

        Parameters
        ----------
        flowchart: seamm.Node
            A non-graphical SEAMM node

        **kwargs : keyword arguments
            Various keyword arguments such as title, namespace or
            extension representing the title displayed in the flowchart,
            the namespace for the plugins of a subflowchart and
            the extension, respectively.

        Returns
        -------
        Energy
        """

        return fhi_aims_step.Energy(flowchart=flowchart, **kwargs)

    def create_tk_node(self, canvas=None, **kwargs):
        """Create and return the graphical Tk node object.

        Parameters
        ----------
        canvas : tk.Canvas
            The Tk Canvas widget

        **kwargs : keyword arguments
            Various keyword arguments such as tk_flowchart, node, x, y, w, h
            representing a graphical flowchart object, a non-graphical node for
            a step, and dimensions of the graphical node.

        Returns
        -------
        TkEnergy
        """

        return fhi_aims_step.TkEnergy(canvas=canvas, **kwargs)

    def description(self):
        """Return a description of what this step does.

        Returns
        -------
        description : dict(str, str)
        """
        return EnergyStep.my_description



# -*- coding: utf-8 -*-

"""Non-graphical part of the FHI-aims step in a SEAMM flowchart
"""

import logging
from pathlib import Path
import pkg_resources
import pprint  # noqa: F401
import sys

import fhi_aims_step
import molsystem
import seamm
from seamm_util import ureg, Q_, getParser  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("FHI-aims")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)


class FHIaims(seamm.Node):
    """
    The non-graphical part of a FHI-aims step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : FHIaimsParameters
        The control parameters for FHI-aims.

    See Also
    --------
    TkFHIaims,
    FHIaims, FHIaimsParameters
    """

    def __init__(
        self,
        flowchart=None,
        title="FHI-aims",
        namespace="org.molssi.seamm.fhi_aims",
        extension=None,
        logger=logger,
    ):
        """A step for FHI-aims in a SEAMM flowchart.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        namespace : str
            The namespace for the plug-ins of the subflowchart
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating FHI-aims {self}")
        self.subflowchart = seamm.Flowchart(
            parent=self, name="FHI-aims", namespace=namespace
        )

        super().__init__(
            flowchart=flowchart,
            title="FHI-aims",
            extension=extension,
            module=__name__,
            logger=logger,
        )
        self.parameters = fhi_aims_step.FHIaimsParameters()

        self._metadata = fhi_aims_step.metadata

    @property
    def version(self):
        """The semantic version of this module."""
        return fhi_aims_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return fhi_aims_step.__git_revision__

    def analyze(self, indent="", **kwargs):
        """Do any analysis of the output from this step.

        Also print important results to the local step.out file using
        "printer".

        Parameters
        ----------
        indent: str
            An extra indentation for the output
        """
        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        # Loop over the subnodes, asking them to do their analysis
        while node is not None:
            for value in node.description:
                printer.important(value)
                printer.important(" ")

            node.analyze()

            node = node.next()

    def create_parser(self):
        """Setup the command-line / config file parser"""
        parser_name = self.step_type
        parser = getParser()

        # Remember if the parser exists ... this type of step may have been
        # found before
        parser_exists = parser.exists(parser_name)

        # Create the standard options, e.g. log-level
        result = super().create_parser(name=parser_name)

        if parser_exists:
            return result

        # FHI-aims specific options

        parser.add_argument(
            parser_name,
            "--max-atoms-to-print",
            default=25,
            help="Maximum number of atoms to print charges, etc.",
        )

        parser.add_argument(
            parser_name,
            "--ncores",
            default="available",
            help=(
                "The maximum number of cores to use for FHI-aims. "
                "Default: all available cores."
            ),
        )
        parser.add_argument(
            parser_name,
            "--atoms-per-core",
            type=int,
            default="5",
            help="the optimal number of atoms per core for FHI-aims",
        )
        parser.add_argument(
            parser_name,
            "--html",
            action="store_true",
            help="whether to write out html files for graphs, etc.",
        )
        parser.add_argument(
            parser_name,
            "--modules",
            nargs="*",
            default=None,
            help="the environment modules to load for FHI-aims",
        )
        parser.add_argument(
            parser_name,
            "--path",
            default=None,
            help="the path to the FHI-aims executables",
        )
        parser.add_argument(
            parser_name,
            "--basis-path",
            default=None,
            help="the path to the basis sets",
        )
        parser.add_argument(
            parser_name,
            "--aims",
            default="aims.x",
            help="the executable for FHI-aims",
        )
        parser.add_argument(
            parser_name, "--mpiexec", default="mpiexec", help="the mpi executable"
        )

        return result

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        self.subflowchart.root_directory = self.flowchart.root_directory

        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        text = self.header + "\n\n"
        while node is not None:
            try:
                text += __(node.description_text(), indent=3 * " ").__str__()
            except Exception as e:
                print(f"Error describing fhi_aims flowchart: {e} in {node}")
                logger.critical(f"Error describing fhi_aims flowchart: {e} in {node}")
                raise
            except:  # noqa: E722
                print(
                    "Unexpected error describing fhi_aims flowchart: "
                    f"{sys.exc_info()[0]} in {str(node)}"
                )
                logger.critical(
                    "Unexpected error describing fhi_aims flowchart: "
                    f"{sys.exc_info()[0]} in {str(node)}"
                )
                raise
            text += "\n"
            node = node.next()

        return text

    def run(self):
        """Run a FHI-aims step.

        Parameters
        ----------
        None

        Returns
        -------
        seamm.Node
            The next node object in the flowchart.
        """
        next_node = super().run(printer)

        # Print our header to the main output
        printer.important(self.header)
        printer.important("")

        # Add the main citation for FHI-aims
        self.references.cite(
            raw=self._bibliography["BLUM20092175"],
            alias="FHI-aims",
            module="FHI-aims step",
            level=1,
            note="The principle citation for FHI-aims.",
        )

        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        # And loop
        while node is not None:
            if node.is_runable:
                node.run()
            node = node.next()

        # Add other citations here or in the appropriate place in the code.
        # Add the bibtex to data/references.bib, and add a self.reference.cite
        # similar to the above to actually add the citation to the references.

        return next_node

    def set_id(self, node_id):
        """Set the id for node to a given tuple"""
        self._id = node_id

        # and set our subnodes
        self.subflowchart.set_ids(self._id)

        return self.next()



# -*- coding: utf-8 -*-

"""Non-graphical part of the Energy step in a FHI-aims flowchart
"""

import csv
import logging
from pathlib import Path
import pkg_resources
import pprint  # noqa: F401
import textwrap

import numpy as np
from tabulate import tabulate

import fhi_aims_step
from .substep import Substep
import molsystem
import seamm
from seamm_util import ureg, Q_  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("FHI-aims")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)


class Energy(Substep):
    """
    The non-graphical part of a Energy step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : EnergyParameters
        The control parameters for Energy.

    See Also
    --------
    TkEnergy,
    Energy, EnergyParameters
    """

    def __init__(
        self,
        flowchart=None,
        title="Energy",
        extension=None,
        module=__name__,
        logger=logger,
    ):
        """A substep for Energy in a subflowchart for FHI-aims.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating Energy {self}")

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            module=__name__,
            logger=logger,
        )

        self._calculation = "energy"
        self._model = None
        self._metadata = fhi_aims_step.metadata
        self.parameters = fhi_aims_step.EnergyParameters()

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return fhi_aims_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return fhi_aims_step.__git_revision__

    def analyze(self, P=None, indent="", configuration=None, data={}, **kwargs):
        """Do any analysis of the output from this step.

        Also print important results to the local step.out file using
        "printer".

        Parameters
        ----------
        indent: str
            An extra indentation for the output
        """
        options = self.parent.options
        directory = Path(self.directory)

        if P is None:
            P = self.parameters.current_values_to_dict(
                context=seamm.flowchart_variables._data
            )

        if len(data) == 0:
            data = self.parse_data()

        text = ""
        if type(self) is fhi_aims_step.Energy:
            # Check that the job ended successfully
            if "is_a_nice_day" in data and data["is_a_nice_day"]:
                text += "The calculation finished successfully."
                # Write a small file to say that LAMMPS ran successfully, so cancel
                # skip if rerunning.
                path = Path(self.directory) / "success.dat"
                path.write_text("success")
            else:
                text += "The calculation did not complete properly! Be cautious.\n\n"

        if len(text) > 0:
            printer.normal(
                __(
                    text,
                    indent=self.indent + 4 * " ",
                    wrap=True,
                    dedent=False,
                )
            )
        text = ""

        # Put any requested results into variables or tables
        self.store_results(configuration=configuration, data=data)

        # Prepare our summary
        table = {
            "Item": [],
            "Value": [],
            "Units": [],
        }
        if "total_energy" in data:
            table["Item"].append("Total energy")
            table["Value"].append(f"{data['total_energy']:.5f}")
            table["Units"].append("eV")
        if "dispersion_energy" in data:
            table["Item"].append("vdW correction")
            table["Value"].append(f"{data['dispersion_energy']:.5f}")
            table["Units"].append("eV")
        if "total_spin" in data:
            table["Item"].append("Total Spin")
            table["Value"].append(f"{data['total_spin']:.5f}")
            table["Units"].append(" ")
        if "relaxation_step_number" in data:
            table["Item"].append("Optimization steps")
            table["Value"].append(data["relaxation_step_number"])
            table["Units"].append(" ")
        if "norm_force_atoms" in data:
            table["Item"].append("Force norm")
            table["Value"].append(data["norm_force_atoms"])
            table["Units"].append("eV/Å")
        if "maximum_force" in data:
            table["Item"].append("Maximum force")
            table["Value"].append(data["maximum_force"])
            table["Units"].append("eV/Å")
        if "total_number_of_loops" in data:
            table["Item"].append("SCF steps")
            table["Value"].append(data["total_number_of_loops"])
            table["Units"].append(" ")
        if "change_charge_density" in data:
            table["Item"].append("Δ charge density")
            table["Value"].append(f"{data['change_charge_density'][-1]:.2e}")
            table["Units"].append(" ")
        if "change_spin_density" in data:
            table["Item"].append("Δ spin density")
            table["Value"].append(f"{data['change_spin_density'][-1]:.2e}")
            table["Units"].append(" ")
        if "change_sum_eigenvalues" in data:
            table["Item"].append("Δ Σ(eigenvalues)")
            table["Value"].append(f"{data['change_sum_eigenvalues'][-1]:.2e}")
            table["Units"].append(" ")
        if "change_forces" in data:
            table["Item"].append("Δ forces")
            table["Value"].append(f"{data['change_forces'][-1]:.2e}")
            table["Units"].append(" ")

        tmp = tabulate(
            table,
            headers="keys",
            tablefmt="psql",
            colalign=("center", "decimal"),
            disable_numparse=True,
        )
        length = len(tmp.splitlines()[0])
        text_lines = []
        text_lines.append("SCF Results".center(length))
        text_lines.append(tmp)
        tmp = "\n\n"
        tmp += textwrap.indent("\n".join(text_lines), self.indent + 7 * " ")
        printer.normal(tmp)

        system, configuration = self.get_system_configuration(None)
        symbols = configuration.atoms.asymmetric_symbols
        atoms = configuration.atoms
        symmetry = configuration.symmetry
        # Charge and spin, if available
        if "atoms_proj_charge" in data:
            # Add to atoms (in coordinate table)
            if "charge" not in atoms:
                atoms.add_attribute(
                    "charge", coltype="float", configuration_dependent=True
                )
            chgs = data["atoms_proj_charge"][0]
            if symmetry.n_symops > 1:
                if P["primitive cell"]:
                    chgs = [chgs[i] for i in self._mapping_to_primitive]
                    print("Unravelling primitive cell")
                    print(f"{chgs=}")
                print(f"{configuration=}, {configuration.id=}")
                print(f"{symmetry=}, {symmetry.id=}")
                chgs, delta = symmetry.symmetrize_atomic_scalar(chgs)
                print(f"{chgs=}")
                delta = np.array(delta)
                max_delta = np.max(abs(delta))
                text += (
                    "The maximum difference of the charges of symmetry related atoms "
                    f"was {max_delta:.4f}\n"
                )
            atoms["charge"][0:] = chgs

            # Print the charges and dump to a csv file
            chg_tbl = {
                "Atom": [*range(1, len(symbols) + 1)],
                "Element": symbols,
                "Charge": [],
            }
            with open(directory / "atom_properties.csv", "w", newline="") as fd:
                writer = csv.writer(fd)
                if "atoms_proj_spin" in data:
                    # Add to atoms (in coordinate table)
                    if "spin" not in atoms:
                        atoms.add_attribute(
                            "spin", coltype="float", configuration_dependent=True
                        )
                        spins = data["atoms_proj_spin"][0]
                        if symmetry.n_symops > 1:
                            if P["primitive cell"]:
                                spins = [spins[i] for i in self._mapping_to_primitive]
                            spins, delta = symmetry.symmetrize_atomic_scalar(spins)
                            atoms["spins"][0:] = spins
                            delta = np.array(delta)
                            max_delta = np.max(abs(delta))
                            text += (
                                " The maximum difference of the spins of symmetry "
                                f"related atoms was {max_delta:.4f}.\n"
                            )
                        atoms["spin"][0:] = spins

                    header = "        Atomic charges and spins"
                    chg_tbl["Spin"] = []
                    writer.writerow(["Atom", "Element", "Charge", "Spin"])
                    for atom, symbol, q, s in zip(
                        range(1, len(symbols) + 1),
                        symbols,
                        chgs,
                        spins,
                    ):
                        q = f"{q:.3f}"
                        s = f"{s:.3f}"

                        writer.writerow([atom, symbol, q, s])

                        chg_tbl["Charge"].append(q)
                        chg_tbl["Spin"].append(s)
                else:
                    header = "        Atomic charges"
                    writer.writerow(["Atom", "Element", "Charge"])
                    for atom, symbol, q in zip(
                        range(1, len(symbols) + 1),
                        symbols,
                        chgs,
                    ):
                        q = f"{q:.2f}"
                        writer.writerow([atom, symbol, q])

                        chg_tbl["Charge"].append(q)
            if len(symbols) <= int(options["max_atoms_to_print"]):
                tmp = tabulate(
                    chg_tbl,
                    headers="keys",
                    tablefmt="psql",
                    colalign=("center", "center"),
                )
                length = len(tmp.splitlines()[0])
                text_lines = []
                text_lines.append(header.center(length))
                text_lines.append(tmp)
                tmp = "\n\n"
                tmp += textwrap.indent("\n".join(text_lines), self.indent + 7 * " ")
                printer.normal(tmp)

        if len(text) > 0:
            printer.normal(
                __(
                    text,
                    indent=self.indent + 4 * " ",
                    wrap=True,
                    dedent=False,
                )
            )
        text = []

        printer.normal("")

    def description_text(self, P=None, configuration=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        if not P:
            P = self.parameters.values_to_dict()

        model = P["model"]
        submodel = P["submodel"]
        if type(self) is fhi_aims_step.Energy:
            text = "A single point"
        else:
            text = "Using an"
        if self.is_expr(model):
            if self.is_expr(submodel):
                text += (
                    " energy calculation with the model and submodel "
                    f"determined at runtime from '{model}' and '{submodel}.'"
                )
        else:
            model = model[0].lower() + model[1:]
            if self.is_expr(submodel):
                text += (
                    f" energy calculation with the {model} and "
                    f"submodel determined at runtime by '{submodel}.'"
                )
            else:
                text += (
                    f" energy calculation with the '{submodel}' variant "
                    f"of the {model}."
                )

        text += " The basis set version is "
        basis = P["basis_version"]
        if self.is_expr(basis):
            text += f"determined by the variable '{basis}'."
        else:
            text += f"'{basis}'"
            level = P[basis + "_level"]
            if self.is_expr(level):
                text += f". The level of the basis set will be determined by '{level}'."
            else:
                text += f", level '{level}'."

        relativity = P["relativity"]
        if self.is_expr(relativity):
            text += (
                " Whether and how relativity will be included will be determined by "
                f"'{relativity}'."
            )
        elif relativity == "none":
            text += " Relativistic effects will not be included."
        else:
            text += f" Relativistic effects will be included using the {relativity}."

        dispersion = P["dispersion"]
        if self.is_expr(dispersion):
            text += (
                " Whether and how long-range van der Waals effects (many-body "
                f"dispersion) will be included will be determined by '{dispersion}'."
            )
        elif dispersion == "none":
            text += " Long-range van der Waals terms will not be included."
        else:
            text += (
                f" Long-range van der Waals terms will be included using {dispersion}."
            )

        if configuration is not None and configuration.periodicity != 0:
            # k-space integration
            kmethod = P["k-grid method"]
            if kmethod == "grid spacing":
                if isinstance(P["odd grid"], bool) or P["odd grid"] == "yes":
                    text += (
                        f" For periodic systems a {P['centering']} grid with a spacing "
                        f"of {P['k-spacing']} and odd numbers of points will be used."
                    )
                else:
                    text += (
                        f" For periodic systems a {P['centering']} grid with a spacing "
                        f"of {P['k-spacing']} will be used."
                    )
            elif kmethod == "explicit grid dimensions":
                text += (
                    f" For periodic systems a {P['na']} x{P['nb']} x{P['nc']} "
                    "grid will be used."
                )
            broadening = P["occupation type"]
            if broadening in ("integer",):
                text += " The occupation will be constrained to be integers."
            else:
                text += (
                    " The effect of temperature on the electron distribution "
                    f"will be modeled using {P['occupation type']} broadening with a "
                    f"width of {P['smearing width']}."
                )
        forces = P["calculate_gradients"]
        if not isinstance(forces, bool) and self.is_expr(forces):
            text += (
                " Whether forces will be calculated in this single-point calculation "
                f"will be determined by '{forces}'."
            )
        if forces == "yes" or (isinstance(forces, bool) and forces):
            text += " The forces will be calculated for this single-point calculation."

        if (
            isinstance(P["input only"], bool)
            and P["input only"]
            or P["input only"] == "yes"
        ):
            text += "\n\nThe input file will be written. No calculation will be run."

        return self.header + "\n" + __(text, indent=4 * " ").__str__()

    def plot_control(self, P, configuration):
        """Create the density and orbital plots if requested

        Parameters
        ----------
        data : dict()
             Dictionary of results from the calculation (results.tag file)
        """
        lines = []

        if P["total density"]:
            lines.append("output cube               total_density")
        if self.is_spin_polarized and P["total spin density"]:
            lines.append("output cube               spin_density")
        if P["difference density"]:
            lines.append("output cube               delta_density")

        # # and work out the orbitals
        # txt = P["selected orbitals"]
        # if txt == "all":
        #     options["PlottedLevels"] = "1:-1"
        # else:
        #     orbitals = []
        #     for chunk in txt.split(","):
        #         chunk = chunk.strip()
        #         if ":" in chunk or ".." in chunk:
        #             if ":" in chunk:
        #                 first, last = chunk.split(":")
        #             elif ".." in chunk:
        #                 first, last = chunk.split("..")
        #             first = first.strip().upper()
        #             last = last.strip().upper()

        #             if first == "HOMO":
        #                 first = homo
        #             elif first == "LUMO":
        #                 first = homo + 1
        #             else:
        #                 first = int(first.removeprefix("HOMO").removeprefix("LUMO"))
        #                 if first < 0:
        #                     first = homo + first
        #                 else:
        #                     first = homo + 1 + first

        #             if last == "HOMO":
        #                 last = homo
        #             elif last == "LUMO":
        #                 last = homo + 1
        #             else:
        #                 last = int(last.removeprefix("HOMO").removeprefix("LUMO"))
        #                 if last < 0:
        #                     last = homo + last
        #                 else:
        #                     last = homo + 1 + last

        #             orbitals.extend(range(first, last + 1))
        #         else:
        #             first = chunk.strip().upper()

        #             if first == "HOMO":
        #                 first = homo
        #             elif first == "LUMO":
        #                 first = homo + 1
        #             else:
        #                 first = int(first.removeprefix("HOMO").removeprefix("LUMO"))
        #                 if first < 0:
        #                     first = homo + first
        #                 else:
        #                     first = homo + 1 + first
        #             orbitals.append(first)

        #         # Remove orbitals out of limits
        #         tmp = orbitals
        #         orbitals = []
        #         for x in tmp:
        #             if x > 0 and x <= n_orbitals:
        #                 orbitals.append(x)

        #         options["PlottedLevels"] = orbitals

        # if periodicity != 0:
        #     if P["selected k-points"] == "all":
        #         options["PlottedKPoints"] = "1:-1"
        #     else:
        #         kpoints = []
        #         for chunk in P["selected k-points"].split(","):
        #             chunk = chunk.strip()
        #             if ":" in chunk or ".." in chunk:
        #                 if ":" in chunk:
        #                     first, last = chunk.split(":")
        #                 elif ".." in chunk:
        #                     first, last = chunk.split("..")
        #                 first = int(first.strip())
        #                 last = int(last.strip())

        #                 if first < 1:
        #                     first = 1

        #                 if last > last_kpoint:
        #                     last = last_kpoint

        #                 kpoints.extend(range(first, last + 1))
        #             else:
        #                 first = int(chunk.strip())
        #                 if first > 0 and first <= last_kpoint:
        #                     kpoints.append(first)
        #         options["PlottedKPoints"] = kpoints

        lines.append("")
        return lines

    def run(self, printer=printer, lines=None):
        """Get the input for an energy step.

        Parameters
        ----------
        None

        Returns
        -------
        str
            The input for FHI-aims
        """
        if lines is None:
            lines = []

        # Get the values of the parameters, dereferencing any variables
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        _, configuration = self.get_system_configuration()

        # Set the attribute for writing just the input
        self.input_only = P["input only"]

        # Print what we are doing
        printer.important(
            __(
                self.description_text(P, configuration=configuration),
                indent=self.indent,
            )
        )

        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        model = P["submodel"]
        if " : " in model:
            model = model.split(" : ")[0].strip()
        lines.append("output                   json_log")
        lines.append(f"xc                       {model}")
        lines.append(f"charge                   {configuration.charge}")

        # Handle spin
        multiplicity = configuration.spin_multiplicity
        self.is_spin_polarized = False
        if P["spin_polarization"] == "none":
            lines.append("spin                      none")
        else:
            atoms = configuration.atoms
            have_spins = "spin" in atoms
            if have_spins:
                for tmp in atoms["spin"]:
                    if tmp is None:
                        have_spins = False
                        break
            if have_spins:
                self.is_spin_polarized = True
                lines.append("spin                     collinear")
                if P["fixed_spin_moment"] == "yes":
                    lines.append(f"fixed_spin_moment        {multiplicity-1}")
                elif isinstance(P["fixed_spin_moment"], int):
                    lines.append(f"fixed_spin_moment        {P['fixed_spin_moment']-1}")
            else:
                lines.append("spin                     none")
        lines.append("output                   Mulliken")

        if P["relativity"] == "atomic ZORA approximation":
            lines.append("relativistic             atomic_zora scalar")
        else:
            lines.append("relativistic             none")
        if P["calculate_gradients"]:
            lines.append("compute_forces           .true.")
        if "MBD-NL" in P["dispersion"]:
            lines.append("many_body_dispersion_nl")
        elif "MBD@rsSCS" in P["dispersion"]:
            lines.append("many_body_dispersion")

        if configuration.periodicity != 0:
            if "explicit" in P["k-grid method"]:
                lines.append(f"k_grid                   {P['na']} {P['nb']} {P['nc']}")
            else:
                lengths = configuration.cell.reciprocal_lengths()
                spacing = P["k-spacing"].to("1/Å").magnitude
                na = round(lengths[0] / spacing)
                nb = round(lengths[0] / spacing)
                nc = round(lengths[0] / spacing)
                na = na if na > 0 else 1
                nb = nb if nb > 0 else 1
                nc = nc if nc > 0 else 1
                if P["odd grid"]:
                    na = na + 1 if na % 2 == 0 else na
                    nb = nb + 1 if nb % 2 == 0 else nb
                    nc = nc + 1 if nc % 2 == 0 else nc

                lines.append(f"k_grid                   {P['na']} {P['nb']} {P['nc']}")

                if P["centering"] == "𝚪-centered":
                    oa = 0.0 if na % 2 == 1 else 1 / (2 * na)
                    ob = 0.0 if nb % 2 == 1 else 1 / (2 * nb)
                    oc = 0.0 if nc % 2 == 1 else 1 / (2 * nc)
                elif P["centering"] == "off center":
                    oa = 0.0 if na % 2 == 0 else 1 / (2 * na)
                    ob = 0.0 if nb % 2 == 0 else 1 / (2 * nb)
                    oc = 0.0 if nc % 2 == 0 else 1 / (2 * nc)
                elif P["centering"] == "Monkhorst-Pack":
                    oa = 0.0 if na % 2 == 1 else 0.5 - 1 / (2 * na)
                    ob = 0.0 if nb % 2 == 1 else 0.5 - 1 / (2 * nb)
                    oc = 0.0 if nc % 2 == 1 else 0.5 - 1 / (2 * nc)
                else:
                    raise RuntimeError(
                        f"Don't recognize k-space grid centering {P['centering']}"
                    )
                lines.append(f"k_offset                 {oa:8.6f} {ob:8.6f} {oc:8.6f}")

                lines.append(
                    f"occupation_type          {P['occupation type'].lower()} "
                    f"{P['smearing width'].m_as('eV'):.4f}"
                )

        # Any plotting we need to do (orbitals, density, ...)
        lines.extend(self.plot_control(P, configuration))

        data = "\n".join(lines)
        data += self._basis_sets(
            P, configuration.atoms.symbols, configuration.atoms.atomic_numbers
        )

        files = {
            "control.in": data,
            "geometry.in": self._geometry(configuration),
        }

        return_files = ["aims.out", "geometry.in.next_step", "*.cube", "*.json"]
        self.run_aims(files, return_files)

        if not self.input_only:
            self.analyze(P=P, configuration=configuration)

            path = directory / "success.dat"
            if not path.exists():
                path.write_text("success")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `fhi_aims_step` package."""

import pytest  # noqa: F401
import fhi_aims_step  # noqa: F401


def test_construction():
    """Just create an object and test its type."""
    result = fhi_aims_step.FHIaims()
    assert str(type(result)) == "<class 'fhi_aims_step.fhi_aims.FHIaims'>"



# -*- coding: utf-8 -*-

"""Base class for substeps in the FHI-aims step
"""

import configparser
import importlib
import json
import logging
from pathlib import Path
import pprint
import re
import shutil

from molsystem.elements import to_symbols
import seamm
import seamm_exec
from seamm_util import Q_, Configuration
import seamm_util.printing as printing

# In addition to the normal logger, two logger-like printing facilities are
# defined: 'job' and 'printer'. 'job' send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# 'printer' sends output to the file 'step.out' in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("FHI-aims")


class Substep(seamm.Node):
    """A base class for substeps in the FHI-aims step."""

    def __init__(
        self,
        flowchart=None,
        title="no title",
        extension=None,
        logger=logger,
        module=__name__,
    ):
        self.mapping_from_primitive = None
        self.mapping_to_primitive = None
        self._is_spin_polarized = False
        self._mapping_from_primitive = None
        self._mapping_to_primitive = None
        self._input_only = False

        super().__init__(
            flowchart=flowchart, title=title, extension=extension, logger=logger
        )

    @property
    def global_options(self):
        return self.parent.global_options

    @property
    def input_only(self):
        """Whether to write the input only, not run MOPAC."""
        return self._input_only

    @input_only.setter
    def input_only(self, value):
        self._input_only = value

    @property
    def is_spin_polarized(self):
        """If the calculation is set up a spin polarized."""
        return self._is_spin_polarized

    @is_spin_polarized.setter
    def is_spin_polarized(self, value):
        self._is_spin_polarized = value

    @property
    def is_runable(self):
        """Indicate whether this not runs or just adds input."""
        return True

    @property
    def options(self):
        return self.parent.options

    def run_aims(self, files, return_files=[]):
        """Run a FHI-aims calculation.

        Parameters
        ----------
        files : {}
            Dictionary of input and control files for FHI-aims
        return_files : [str]
            List of files to bring backfrom the calculation

        Returns
        -------
        """
        directory = Path(self.directory)
        if self.input_only:
            for filename in files:
                path = directory / filename
                path.write_text(files[filename])
            return

        # Check for successful run, don't rerun
        path = directory / "success.dat"
        if path.exists():
            result = {}
            path = directory / "stdout.txt"
            if path.exists():
                result["stdout"] = path.read_text()
            result["stderr"] = ""
            return result

        # Access the options
        options = self.options
        seamm_options = self.global_options

        # Get the computational environment and set limits
        ce = seamm_exec.computational_environment()

        atoms_per_core = options["atoms_per_core"]
        max_cores = options["ncores"]
        if max_cores == "available":
            max_cores = self.global_options["ncores"]

        # How many processors does this node have?
        n_cores = ce["NTASKS"]
        self.logger.info("The number of cores is {}".format(n_cores))
        if max_cores == "available":
            max_cores = n_cores
        else:
            max_cores = int(max_cores)

        # Get the current configuration because it might change between steps
        _, configuration = self.get_system_configuration()
        n_atoms = configuration.n_atoms
        np = n_atoms // atoms_per_core
        if np <= 0:
            np = 1
        elif np > max_cores:
            np = max_cores

        ce["NTASKS"] = np

        printer.normal(self.indent + 4 * " " + f"FHI-aims running using {np} cores.")

        cmd = ["{code}"]

        # Run FHI-aims
        executor = self.parent.flowchart.executor

        # Read configuration file for FHI-aims
        executor_type = executor.name
        full_config = configparser.ConfigParser()
        ini_dir = Path(seamm_options["root"]).expanduser()
        path = ini_dir / "fhi-aims.ini"

        # If the config file doesn't exists, get the default
        if not path.exists():
            resources = importlib.resources.files("fhi-aims_step") / "data"
            ini_text = (resources / "fhi-aims.ini").read_text()
            txt_config = Configuration(path)
            txt_config.from_string(ini_text)
            txt_config.save()

        full_config.read(ini_dir / "fhi-aims.ini")

        # Getting desperate! Look for an executable in the path
        if executor_type not in full_config:
            path = shutil.which("fhi-aims")
            if path is None:
                raise RuntimeError(
                    f"No section for '{executor_type}' in FHI-aims ini file "
                    f"({ini_dir / 'fhi-aims.ini'}), nor in the defaults, nor "
                    "in the path!"
                )
            else:
                txt_config = Configuration(path)
                txt_config.add_section(executor_type)
                txt_config.set_value(executor_type, "installation", "local")
                txt_config.set_value(executor_type, "fhi-aims", str(path))
                path = shutil.which("mpiexec")
                if path is not None:
                    txt_config.set_value(executor_type, "mpiexec", str(path))
                txt_config.save()
                full_config.read(ini_dir / "fhi-aims.ini")

        config = dict(full_config.items(executor_type))

        result = executor.run(
            cmd=cmd,
            ce=ce,
            config=config,
            directory=self.directory,
            files=files,
            return_files=return_files,
            in_situ=True,
            shell=True,
            env={"OMP_NUM_THREADS": "1"},
        )

        if not result:
            self.logger.error("There was an error running FHI-aims")
            return None

        logger.debug("\n" + pprint.pformat(result))

        return result

    def _basis_sets(self, P, symbols, atnos):
        """The basis set part of the control file.

        Parameters
        ----------
        P : {str: value}
            The options for this substep
        symbols : [str]
            The symbols of the elements
        atnos : [int]
            The atomic numbers of the elements

        Returns
        -------
        str
            The basis set text for FHI-aims
        """
        # Get the configuration for FHI-aims to find where the basis sets are.
        executor = self.parent.flowchart.executor

        # Read configuration file for FHI-aims
        ini_dir = Path(self.global_options["root"]).expanduser()
        full_config = configparser.ConfigParser()
        full_config.read(ini_dir / "fhi-aims.ini")
        executor_type = executor.name
        if executor_type not in full_config:
            raise RuntimeError(
                f"No section for '{executor_type}' in the FHI-aims ini file "
                f"({ini_dir / 'fhi-aims.ini'})"
            )
        config = dict(full_config.items(executor_type))

        basis = P["basis_version"]
        level = P[basis + "_level"]
        basis_path = Path(config["basis-path"]) / basis / level

        # The model chemistry
        model = P["submodel"]
        if " : " in model:
            model = model.split(" : ")[0].strip()
        self.model = f"{model}/{basis}({level})"

        filename = {
            atno: f"{atno:02}_{symbol}_default" for atno, symbol in zip(atnos, symbols)
        }
        result = ""
        for atno in sorted(filename.keys()):
            path = basis_path / filename[atno]
            result += path.read_text()
        return result

    def _geometry(self, configuration):
        """Get the geometry for FHI-aims.

        Parameters
        ----------
        configuration : molsystem._Configuration
            The current configuration

        Returns
        -------
        str
            The contents of the geometry.in for FHI-aims
        """

        key = "atom" if configuration.periodicity == 0 else "atom_frac"

        lines = []
        lines.append(f"# Geometry for {configuration.system.name}/{configuration.name}")
        lines.append("#")
        if configuration.periodicity == 0:
            xyzs = configuration.coordinates
            symbols = configuration.atoms.symbols
        else:
            if "primitive cell" in self.parameters:
                use_primitive_cell = self.parameters["primitive cell"].get(
                    context=seamm.flowchart_variables._data
                )
            else:
                use_primitive_cell = True

            if use_primitive_cell:
                # Write the structure using the primitive cell
                (
                    lattice,
                    xyzs,
                    atomic_numbers,
                    self._mapping_from_primitive,
                    self._mapping_to_primitive,
                ) = configuration.primitive_cell()
                lines.append("# Using primitive cell. Conventional cell is")
            else:
                # Use the full cell
                lattice = configuration.cell.vectors()
                xyzs = configuration.atoms.get_coordinates(fractionals=True)
                atomic_numbers = configuration.atoms.atomic_numbers

                n_atoms = len(atomic_numbers)
                self._mapping_from_primitive = [i for i in range(n_atoms)]
                self._mapping_to_primitive = [i for i in range(n_atoms)]
            symbols = to_symbols(atomic_numbers)

            a, b, c, alpha, beta, gamma = configuration.cell.parameters
            lines.append(
                f"# cell = {a:4f} {b:.4f} {c:.4f} {alpha:.2f} {beta:.2f} {gamma:.2f}"
            )
            for abc in lattice:
                x, y, z = abc
                lines.append(f"lattice_vector  {x:15.9f} {y:15.9f} {z:15.9f}")
            lines.append("#")

        if self.is_spin_polarized:
            spins = configuration.atoms["spin"]
            for xyz, symbol, spin in zip(xyzs, symbols, spins):
                x, y, z = xyz
                lines.append(f"{key}  {x:15.8f}  {y:15.8f}  {z:15.8f} {symbol}")
                lines.append(f"initial_moment  {spin:8.4f}")
        else:
            for xyz, symbol in zip(xyzs, symbols):
                x, y, z = xyz
                lines.append(f"{key}  {x:15.8f}  {y:15.8f}  {z:15.8f} {symbol}")
        return "\n".join(lines)

    def parse_data(self):
        """Get the data from the AIMS files

        Parameters
        ----------
        none
        """
        data = {}

        # Parsing values from the JSON
        path = Path(self.directory) / "aims.json"
        if path.exists():
            with path.open() as fd:
                jdata = json.load(fd)
        for section in jdata:
            record = section["record_type"]
            for key, mdata in self._metadata["results"].items():
                self.logger.debug(f"results {key=}")
                if (
                    "calculation" in mdata
                    and self.calculation not in mdata["calculation"]
                ):
                    continue
                if "json_section" not in mdata or mdata["json_section"] != record:
                    continue

                tmp_key = key.rstrip("^")
                if tmp_key in section:
                    value = section[tmp_key]
                    if "energy" in tmp_key:
                        # AIMS has a bad conversion to eV. Fix
                        aimsEh2eV = (
                            float(section["hartree"])
                            if "hartree" in section
                            else 27.2113845
                        )
                        value = value / aimsEh2eV * Q_(1, "E_h").m_as("eV")
                    if "type" in mdata:
                        if isinstance(value, list):
                            value = [mdata["type"](v) for v in value]
                        else:
                            value = mdata["type"](value)
                    if mdata["dimensionality"] == "scalar":
                        data[key] = value
                    else:
                        if key not in data:
                            data[key] = []
                        data[key].append(value)

        if "is_a_nice_day" in section:
            data["is_a_nice_day"] = section["is_a_nice_day"]

        # Parsing values from the output
        path = Path(self.directory) / "aims.out"
        output = path.read_text()

        # Find the citations
        lines = output.splitlines()

        # Check that the job ended successfully
        if "is_a_nice_day" not in data:
            if "          Have a nice day." in lines:
                data["is_a_nice_day"] = True
            else:
                data["is_a_nice_day"] = False

        if "    http://doi.org/10.1016/j.cpc.2017.09.007" in lines:
            self.references.cite(
                raw=self._bibliography["YU2018267"],
                alias="ELSI",
                module="FHI-aims step",
                level=1,
                note="The principle citation for ELSI.",
            )
        if "    http://doi.org/10.1016/j.jcp.2009.08.008" in lines:
            self.references.cite(
                raw=self._bibliography["HAVU20098367"],
                alias="Numerical grids for integration",
                module="FHI-aims step",
                level=1,
                note="The principle citation for numerical integration grids.",
            )

        for key, mdata in self._metadata["results"].items():
            self.logger.debug(f"results {key=}")
            if key in data:
                continue
            if "calculation" in mdata and self.calculation not in mdata["calculation"]:
                continue
            if "re" not in mdata:
                continue

            self.logger.debug(mdata["re"])
            matches = re.findall(mdata["re"], output)
            self.logger.debug(matches)
            if len(matches) > 0:
                txt = matches[-1]
                if txt == "":
                    self.logger.warning(
                        f"Parsing output, re ({mdata['re']}) gave error"
                    )
                else:
                    if "type" in mdata:
                        data[key] = mdata["type"](txt)
                    else:
                        data[key] = txt

        return data



# -*- coding: utf-8 -*-

"""Non-graphical part of the Optimization step in a FHI aims flowchart
"""

import logging
from pathlib import Path
import pkg_resources
import pprint  # noqa: F401
import traceback

import fhi_aims_step  # noqa: E999
import molsystem
import seamm
from seamm_util import ureg, Q_  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("FHI aims")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)


class Optimization(fhi_aims_step.Energy):
    """
    The non-graphical part of a Optimization step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : OptimizationParameters
        The control parameters for Optimization.

    See Also
    --------
    TkOptimization,
    Optimization, OptimizationParameters
    """

    def __init__(
        self,
        flowchart=None,
        title="Optimization",
        extension=None,
        module=__name__,
        logger=logger,
    ):
        """A substep for Optimization in a subflowchart for FHI aims.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating Optimization {self}")

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            module=module,
            logger=logger,
        )

        self._calculation = "optimization"
        self._model = None
        self._metadata = fhi_aims_step.metadata
        self.parameters = fhi_aims_step.OptimizationParameters()

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return fhi_aims_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return fhi_aims_step.__git_revision__

    def description_text(self, P=None, configuration=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        if not P:
            P = self.parameters.values_to_dict()

        tmp = super().description_text(P)
        energy_description = "\n".join(tmp.splitlines()[1:])

        text = (
            "The optimization using BFGS will stop at force convergence of "
            f"{P['force_convergence']}."
        )
        if configuration is not None and configuration.periodicity != 0:
            if P["optimize_cell"] == "yes":
                text += " For periodic systems, the unit cell will also be optimized."
            elif "angle" in P["optimize_cell"]:
                text += (
                    " For periodic systems, the unit cell will also be optimized, but "
                    "cell angles will be fixed."
                )
            if P["pressure"] != 0.0:
                text += f" An external pressure of {P['pressure']} will be applied."

        if not (
            isinstance(P["input only"], bool)
            and P["input only"]
            or P["input only"] == "yes"
        ):
            text += f" The optimized structure will {P['structure handling']} "

            confname = P["configuration name"]
            if confname == "use SMILES string":
                text += "using SMILES as its name."
            elif confname == "use Canonical SMILES string":
                text += "using canonical SMILES as its name."
            elif confname == "keep current name":
                text += "keeping the current name."
            elif confname == "optimized with {model}":
                text += "with 'optimized with <model>' as its name."
            elif confname == "use configuration number":
                text += "using the index of the configuration (1, 2, ...) as its name."
            else:
                confname = confname.replace("{model}", "<model>")
                text += f"with '{confname}' as its name."

        return (
            self.header
            + "\n"
            + __(text, indent=4 * " ").__str__()
            + "\n\n"
            + energy_description
        )

    def run(self, lines=None):
        """Run an Optimization step.

        Parameters
        ----------
        None

        Returns
        -------
        seamm.Node
            The next node object in the flowchart.
        """
        if lines is None:
            lines = []

        # Get the values of the parameters, dereferencing any variables
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        # Get the current system and configuration (ignoring the system...)
        _, configuration = self.get_system_configuration(None)

        # Set the attribute for writing just the input
        self.input_only = P["input only"]

        convergence = P["force_convergence"].m_as("eV/Å")
        lines.append(f"relax_geometry           bfgs {convergence:.4f}")

        if configuration.periodicity != 0:
            if P["optimize_cell"] == "yes":
                lines.append("relax_unit_cell          full")
            elif "angle" in P["optimize_cell"]:
                lines.append("relax_unit_cell          fixed_angles")

            if P["pressure"].magnitude != 0.0:
                lines.append(f"external_pressure        {P['pressure'].m_as('eV/Å^3')}")

        super().run(printer=printer, lines=lines)

        # Add other citations here or in the appropriate place in the code.
        # Add the bibtex to data/references.bib, and add a self.reference.cite
        # similar to the above to actually add the citation to the references.

    def analyze(self, P=None, indent="", configuration=None, data={}, **kwargs):
        """Do any analysis of the output from this step.

        Also print important results to the local step.out file using
        "printer".

        Parameters
        ----------
        indent: str
            An extra indentation for the output
        """
        # Get the parameters used
        if P is None:
            P = self.parameters.current_values_to_dict(
                context=seamm.flowchart_variables._data
            )

        if len(data) == 0:
            data = self.parse_data()

        # Check that the job ended successfully
        if "is_a_nice_day" in data and data["is_a_nice_day"]:
            text = "The optimization finished successfully."
            # Write a small file to say that LAMMPS ran successfully, so cancel
            # skip if rerunning.
            path = Path(self.directory) / "success.dat"
            path.write_text("success")
        else:
            text = "The optimization did not complete properly! Be cautious.\n\n"

        path = Path(self.directory) / "geometry.in.next_step"
        coordinate_type = None
        if path.exists():
            lines = path.read_text().splitlines()
            xyz = []
            lattice_vectors = []
            for line in lines:
                line = line.strip()
                if len(line) == 0:
                    continue
                if line[0] == "#":
                    continue
                tmp = line.split()
                if tmp[0] == "atom":
                    coordinate_type = "Cartesian"
                    x, y, z = tmp[1:4]
                    xyz.append([float(x), float(y), float(z)])
                elif tmp[0] == "atom_frac":
                    coordinate_type = "fractional"
                    x, y, z = tmp[1:4]
                    xyz.append([float(x), float(y), float(z)])
                elif tmp[0] == "lattice_vector":
                    x, y, z = tmp[1:4]
                    lattice_vectors.append([float(x), float(y), float(z)])

            # Follow instructions for where to put the coordinates,
            _, starting_configuration = self.get_system_configuration(None)
            system, configuration = self.get_system_configuration(
                P=P, same_as="current", model=self.model
            )

            if configuration.periodicity == 3:
                (
                    lattice_in,
                    fractionals_in,
                    atomic_numbers,
                    self._mapping_from_primitive,
                    self._mapping_to_primitive,
                ) = starting_configuration.primitive_cell()

                print("Updating structure after optimization")
                print(f"{configuration=}, {configuration.id=}")
                symmetry = configuration.symmetry
                print(f"{symmetry=}, {symmetry.id=}")
                print("XYZ")
                pprint.pprint(xyz)
                print("Lattice")
                pprint.pprint(lattice_vectors)

                tmp = configuration.update(
                    xyz,
                    fractionals=coordinate_type == "fractional",
                    atomic_numbers=atomic_numbers,
                    lattice=lattice_vectors,
                    space_group=starting_configuration.symmetry.group,
                    symprec=0.01,
                )

                print("Updated coordinates")
                pprint.pprint(configuration.atoms.get_coordinates(asymmetric=True))
                print("full coordinates")
                pprint.pprint(configuration.atoms.get_coordinates(asymmetric=False))

                # Symmetry may have changed
                if tmp != "":
                    logger.warning(tmp)
                    text += f"\n\nWarning: {tmp}\n\n"
                    (
                        lattice,
                        fractionals,
                        atomic_numbers,
                        self._mapping_from_primitive,
                        self._mapping_to_primitive,
                    ) = configuration.primitive_cell()
            else:
                configuration.atoms.set_coordinates(
                    xyz, fractionals=coordinate_type == "fractional"
                )
        elif "relaxation_step_number" in data and data["relaxation_step_number"] == 0:
            # When aims does not steps it doesn't write a structure file :-)
            # Just copy the previous structure to any new one created...
            _, starting_configuration = self.get_system_configuration(None)
            system, configuration = self.get_system_configuration(
                P=P, same_as="current", model=self.model
            )

        # Write the structure out for viewing.
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        #  MMCIF file has bonds
        try:
            path = directory / "optimized.mmcif"
            path.write_text(configuration.to_mmcif_text())
        except Exception:
            message = "Error creating the mmcif file\n\n" + traceback.format_exc()
            logger.warning(message)
        # CIF file has cell
        if configuration.periodicity == 3:
            try:
                path = directory / "optimized.cif"
                path.write_text(configuration.to_cif_text())
            except Exception:
                message = "Error creating the cif file\n\n" + traceback.format_exc()
                logger.warning(message)

        printer.normal(
            __(
                text,
                indent=self.indent + 4 * " ",
                wrap=True,
                dedent=False,
            )
        )

        super().analyze(indent=indent, configuration=configuration, **kwargs)

        text += (
            f" The optimized structure was placed in configuration {system.name}/"
            f"{configuration.name}."
        )

        printer.normal(
            __(
                text,
                indent=self.indent + 4 * " ",
                wrap=True,
                dedent=False,
            )
        )
        printer.normal("")



# -*- coding: utf-8 -*-

"""The graphical part of a Energy step"""

import logging
import pprint  # noqa: F401
import tkinter as tk
import tkinter.ttk as ttk

import fhi_aims_step  # noqa: F401, E999
import seamm
from seamm_util import ureg, Q_, units_class  # noqa: F401, E999
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkEnergy(seamm.TkNode):
    """
    The graphical part of a Energy step in a flowchart.

    Attributes
    ----------
    tk_flowchart : TkFlowchart = None
        The flowchart that we belong to.
    node : Node = None
        The corresponding node of the non-graphical flowchart
    canvas: tkCanvas = None
        The Tk Canvas to draw on
    dialog : Dialog
        The Pmw dialog object
    x : int = None
        The x-coordinate of the center of the picture of the node
    y : int = None
        The y-coordinate of the center of the picture of the node
    w : int = 200
        The width in pixels of the picture of the node
    h : int = 50
        The height in pixels of the picture of the node
    self[widget] : dict
        A dictionary of tk widgets built using the information
        contained in Energy_parameters.py

    See Also
    --------
    Energy, TkEnergy,
    EnergyParameters,
    """

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """
        Initialize a graphical node.

        Parameters
        ----------
        tk_flowchart: Tk_Flowchart
            The graphical flowchart that we are in.
        node: Node
            The non-graphical node for this step.
        namespace: str
            The stevedore namespace for finding sub-nodes.
        canvas: Canvas
           The Tk canvas to draw on.
        x: float
            The x position of the nodes center on the canvas.
        y: float
            The y position of the nodes cetner on the canvas.
        w: float
            The nodes graphical width, in pixels.
        h: float
            The nodes graphical height, in pixels.

        Returns
        -------
        None
        """
        self.dialog = None

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

    def create_dialog(self, title="Energy"):
        """
        Create the dialog. A set of widgets will be chosen by default
        based on what is specified in the Energy_parameters
        module.

        Parameters
        ----------
        None

        Returns
        -------
        None

        See Also
        --------
        TkEnergy.reset_dialog
        """

        # Shortcut for parameters
        P = self.node.parameters

        frame = super().create_dialog(title=title)

        # Make scrollable in case too large
        self["scrolled frame"] = sw.ScrolledFrame(frame)
        self["scrolled frame"].grid(row=0, column=0, sticky=tk.NSEW)
        frame.rowconfigure(0, weight=1)
        frame.columnconfigure(0, weight=1)

        main_frame = self["main frame"] = self["scrolled frame"].interior()

        # Then create the widgets
        for key in ("input only", "gui", "calculate_gradients"):
            self[key] = P[key].widget(main_frame)

        # Patch width of GUI widget ... it is too wide by default
        self["gui"].configure(width=75)
        self["gui"].combobox.state(["readonly"])

        # Frame to isolate widgets
        e_frame = self["energy frame"] = ttk.LabelFrame(
            main_frame,
            borderwidth=4,
            relief="sunken",
            text="Electronic Structure Definition",
            labelanchor="n",
            padding=10,
        )

        # Then create the widgets
        for key in fhi_aims_step.EnergyParameters.energy_parameters:
            self[key] = P[key].widget(e_frame)
            try:
                self[key].configure(width=45)
            except Exception:
                pass

        for key in ("gui", "basis_version", "model", "submodel"):
            self[key].bind("<<ComboboxSelected>>", self.reset_energy_frame)
            self[key].bind("<Return>", self.reset_energy_frame)
            self[key].bind("<FocusOut>", self.reset_energy_frame)

        # The k-space integration
        k_frame = self["k-space frame"] = ttk.LabelFrame(
            main_frame,
            borderwidth=4,
            relief="sunken",
            text="k-space integration (for periodic systems)",
            labelanchor="n",
            padding=10,
        )
        for key in fhi_aims_step.EnergyParameters.kspace_parameters:
            self[key] = P[key].widget(k_frame)

        for key in ("na", "nb", "nc"):
            self[key].entry.configure(width=4)

        for key in ("k-grid method", "occupation type"):
            self[key].bind("<<ComboboxSelected>>", self.reset_kspace_frame)
            self[key].bind("<Return>", self.reset_kspace_frame)
            self[key].bind("<FocusOut>", self.reset_kspace_frame)

        # A tab for output of orbitals, etc.
        notebook = self["notebook"]
        self["output frame"] = oframe = ttk.Frame(notebook)
        notebook.insert(self["results frame"], oframe, text="Output", sticky="new")

        # Frame to isolate widgets
        p_frame = self["plot frame"] = ttk.LabelFrame(
            self["output frame"],
            borderwidth=4,
            relief="sunken",
            text="Plots",
            labelanchor="n",
            padding=10,
        )

        for key in fhi_aims_step.EnergyParameters.output_parameters:
            self[key] = P[key].widget(p_frame)

        # Set the callbacks for changes
        for widget in ("orbitals",):
            w = self[widget]
            w.combobox.bind("<<ComboboxSelected>>", self.reset_plotting)
            w.combobox.bind("<Return>", self.reset_plotting)
            w.combobox.bind("<FocusOut>", self.reset_plotting)
        p_frame.grid(row=0, column=0, sticky="new")
        oframe.columnconfigure(0, weight=1)

        # and lay them out
        self.reset_plotting()

    def edit(self):
        """Present a dialog for editing this step's parameters.

        Subclasses can override this.
        """
        # Create the dialog if it doesn't exist
        if self.dialog is None:
            self.create_dialog()
            # After full creation, reset the dialog. This may do nothing,
            # or may layout the widgets, but can only be done after fully
            # creating the dialog.
            self.reset_dialog()
            # And resize the dialog, making it large
            screen_w = self.dialog.winfo_screenwidth()
            screen_h = self.dialog.winfo_screenheight()
            w = int(0.95 * screen_w)
            h = int(0.9 * screen_h)
            x = int(0.05 * screen_w / 2)
            y = int(0.1 * screen_h / 2)

            self.dialog.geometry(f"{w}x{h}+{x}+{y}")

        # And put it on-screen, the first time centered. If it contains
        # a subflowchart, save it so it can be restored on a 'Cancel'
        if self.tk_subflowchart is not None:
            self.tk_subflowchart.push()

        self.dialog.activate(geometry="centerscreenfirst")

    def reset_dialog(self, widget=None):
        """Layout the widgets in the dialog.

        The widgets are chosen by default from the information in
        Energy_parameter.

        This function simply lays them out row by row with
        aligned labels. You may wish a more complicated layout that
        is controlled by values of some of the control parameters.
        If so, edit or override this method

        Parameters
        ----------
        widget : Tk Widget = None

        Returns
        -------
        None

        See Also
        --------
        TkEnergy.create_dialog
        """

        # Remove any widgets previously packed
        frame = self["main frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # Shortcut for parameters
        P = self.node.parameters

        # keep track of the row in a variable, so that the layout is flexible
        # if e.g. rows are skipped to control such as "method" here
        row = 0

        # Whether to just write input
        self["input only"].grid(row=row, column=0, sticky=tk.W)
        row += 1

        # The level of the GUI
        self["gui"].grid(row=row, column=0, columnspan=4, pady=5)
        row += 1

        # The model for the calculation
        self["energy frame"].grid(row=row, column=0, columnspan=2, pady=5, sticky=tk.N)
        self["k-space frame"].grid(row=row, column=2, columnspan=2, pady=5, sticky=tk.N)
        row += 1

        # Other parameters for a single-point calculation
        if type(self) is fhi_aims_step.TkEnergy:
            keys = ["calculate_gradients"]
            for key in keys:
                self[key].grid(row=row, column=0, columnspan=4, pady=10)
                row += 1

        # Layout the energy widgets
        self.reset_energy_frame()
        self.reset_kspace_frame()

        # Setup the results if there are any
        have_results = (
            "results" in self.node.metadata and len(self.node.metadata["results"]) > 0
        )
        if have_results and "results" in P:
            self.setup_results()

        return row

    def reset_energy_frame(self, widget=None):
        """Layout the widgets in the dialog.

        The widgets are chosen by default from the information in
        Energy_parameter.

        This function simply lays them out row by row with
        aligned labels. You may wish a more complicated layout that
        is controlled by values of some of the control parameters.
        If so, edit or override this method

        Parameters
        ----------
        widget : Tk Widget = None

        Returns
        -------
        None

        See Also
        --------
        TkEnergy.create_dialog
        """

        # Remove any widgets previously packed
        frame = self["energy frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # Shortcut for parameters
        P = self.node.parameters

        # keep track of the row in a variable, so that the layout is flexible
        # if e.g. rows are skipped to control such as "method" here
        row = 0

        # The level of the gui
        tmp = self["gui"].get()
        if "energy" in tmp:
            if "molecules" in tmp:
                gui = "molecule/energy"
            elif " metallic" in tmp:
                gui = "metal/energy"
            else:
                gui = "standard"
        elif "band" in tmp:
            if "molecules" in tmp:
                gui = "molecule/band"
            elif " metallic" in tmp:
                gui = "metal/band"
            else:
                gui = "standard"
        else:
            gui = "experimental" if "experimental" in tmp else "standard"

        # The model for the calculation
        keys = ["model", "submodel"]
        model = self["model"].get()

        # And what models have available sublevels?
        model_data = self.node.metadata["computational models"][
            "Density Functional Theory (DFT)"
        ]["models"]
        models = []
        for tmp in model_data.keys():
            choices = model_data[tmp]["parameterizations"]
            if gui == "experimental":
                if len(choices) > 0:
                    models.append(tmp)
            elif gui == "standard":
                if (
                    len(
                        [
                            m
                            for m in choices.keys()
                            if "experimental" not in choices[m]["gui"]
                        ]
                    )
                    > 0
                ):
                    models.append(tmp)
            else:
                if len([m for m in choices.keys() if gui in choices[m]["gui"]]) > 0:
                    models.append(tmp)
        w = self["model"]
        if w.combobox.cget("values") != models:
            w.config(values=models)
            if not self.is_expr(model) and model not in models and len(models) > 0:
                w.set(models[0])
            else:
                w.set(model)

        # Fill out the possible submodels
        submodel = self["submodel"].get()
        choices = model_data[model]["parameterizations"]
        if gui == "experimental":
            submodels = [*choices.keys()]
        elif gui == "standard":
            submodels = [
                m for m in choices.keys() if "experimental" not in choices[m]["gui"]
            ]
        else:
            submodels = [m for m in choices.keys() if gui in choices[m]["gui"]]

        w = self["submodel"]
        if w.combobox.cget("values") != submodels:
            w.config(values=submodels)
            if (
                not self.is_expr(submodel)
                and submodel not in submodels
                and len(submodels) > 0
            ):
                w.set(submodels[0])
            else:
                w.set(submodel)

        # And the basis
        widgets = []
        keys.append("basis_version")

        basis_version = self["basis_version"].get()
        key = basis_version + "_level"
        if key in P:
            keys.append(key)
        else:
            tk.messagebox.showwarning(
                title=f"{basis_version} basis not supported yet",
                message=f"{basis_version} basis not supported yet",
            )

        keys.append("spin_polarization")
        keys.append("relativity")
        keys.append("dispersion")
        keys.append("primitive cell")

        for key in keys:
            self[key].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        # Align the labels
        sw.align_labels(widgets, sticky=tk.E)

        # Setup the results if there are any
        have_results = (
            "results" in self.node.metadata and len(self.node.metadata["results"]) > 0
        )
        if have_results and "results" in P:
            self.setup_results()

    def reset_kspace_frame(self, widget=None):
        """Layout the widgets in the dialog.

        The widgets are chosen by default from the information in
        Energy_parameter.

        This function simply lays them out row by row with
        aligned labels. You may wish a more complicated layout that
        is controlled by values of some of the control parameters.
        If so, edit or override this method

        Parameters
        ----------
        widget : Tk Widget = None

        Returns
        -------
        None

        See Also
        --------
        TkEnergy.create_dialog
        """

        # Remove any widgets previously packed
        frame = self["k-space frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # keep track of the row in a variable, so that the layout is flexible
        # if e.g. rows are skipped to control such as "method" here
        row = 0

        # The model for the calculation
        key = "k-grid method"
        method = self[key].get()
        self[key].grid(row=row, column=0, columnspan=4, sticky=tk.W)
        row += 1

        if "explicit" in method:
            for col, key in enumerate(("na", "nb", "nc"), start=1):
                self[key].grid(row=row, column=col, sticky=tk.EW)
            row += 1
        else:
            widgets = []
            for key in ("k-spacing", "odd grid", "centering"):
                self[key].grid(row=row, column=1, columnspan=3, sticky=tk.EW)
                widgets.append(self[key])
                row += 1
            sw.align_labels(widgets, sticky=tk.E)
        frame.columnconfigure(0, minsize=30)

        key = "occupation type"
        smearing = self[key].get()
        self[key].grid(row=row, column=0, columnspan=4, sticky=tk.W)
        row += 1

        if smearing not in ("integer",):
            key = "smearing width"
            self[key].grid(row=row, column=1, columnspan=3, sticky=tk.W)
            row += 1

    def right_click(self, event):
        """
        Handles the right click event on the node.

        Parameters
        ----------
        event : Tk Event

        Returns
        -------
        None

        See Also
        --------
        TkEnergy.edit
        """

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def reset_plotting(self, widget=None):
        frame = self["plot frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        plot_orbitals = self["orbitals"].get() == "yes"

        widgets = []

        row = 0
        for key in (
            "total density",
            "total spin density",
            "difference density",
            "orbitals",
        ):
            self[key].grid(row=row, column=0, columnspan=4, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        if plot_orbitals:
            key = "selected orbitals"
            self[key].grid(row=row, column=1, columnspan=4, sticky=tk.EW)
            row += 1
            key = "selected k-points"
            self[key].grid(row=row, column=1, columnspan=4, sticky=tk.EW)
            row += 1

        key = "region"
        self[key].grid(row=row, column=0, columnspan=4, sticky=tk.EW)
        widgets.append(self[key])
        row += 1

        key = "nx"
        self[key].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
        widgets.append(self[key])
        self["ny"].grid(row=row, column=2, sticky=tk.EW)
        self["nz"].grid(row=row, column=3, sticky=tk.EW)

        sw.align_labels(widgets, sticky=tk.E)
        frame.columnconfigure(0, minsize=10)
        frame.columnconfigure(4, weight=1)

        return row
